/*
 * Project: Spherelib
 * Date: 09/04  
 * Authors: M. Le Jeune
 *
 */

/* Copyright (C) 2008 APC CNRS Université Paris Diderot <lejeune@apc.univ-paris7.fr>  
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see http://www.gnu.org/licenses/gpl.html
 */
#include "CovMat.h"
#include <gsl/gsl_rng.h>
#include <gsl/gsl_sort_double.h>

using namespace std;

/////////////////////////////////////////////////
/// Return a gsl matrix for bin z
/// \param CMat : Input covariance matrices object
/// \param z : bin number
/// \param M : pointor to an allocated gsl matrix
/////////////////////////////////////////////////
template <typename T> void setMatrix (const arr2<T> &CMat, int z, int nbin, gsl_matrix *M)
{
  int k = 0;
  for (int i=0; i<nbin; i++)
    for (int j=0; j<=i; j++)
      {
	gsl_matrix_set (M, i, j, CMat[k][z]);
	gsl_matrix_set (M, j, i, CMat[k++][z]);
      }
}

/////////////////////////////////////////////////
/// compute ILC filter : 
/// \param CMat : Input covariance matrices object
/// \param Filter : array of filter coefficients
/// \param MixCol : mixing column
/////////////////////////////////////////////////
template <typename T> Status ILC (CovMat<T> CMat, arr2<T> &Filter, vector<double> MixCol)
{
  int NbIn = CMat.getNbIn();
  int NbBin = CMat.getNbBin();
  Filter.alloc(NbIn, NbBin);
  gsl_matrix * M, *iM;
  gsl_vector *a, *Ma;
  gsl_permutation *p;
  double aMa;
  M  = gsl_matrix_alloc (NbIn, NbIn);
  iM = gsl_matrix_alloc (NbIn, NbIn);
  a  = gsl_vector_alloc (NbIn);
  Ma = gsl_vector_alloc (NbIn);
  p  = gsl_permutation_alloc (NbIn);
  CBLAS_UPLO_t Uplo=CblasUpper;
  
  for (int i=0; i<NbIn; i++)
    gsl_vector_set(a, i, MixCol[i]);
  
  int signum;
  const arr2<T> CArr = CMat.getCovMat();
    
  for (long z=0; z<NbBin; z++)
    {
      setMatrix(CArr, z, NbIn, M);
      gsl_matrix_set_zero (iM);
      gsl_vector_set_zero (Ma);

      gsl_linalg_LU_decomp (M, p, &signum);
      gsl_linalg_LU_invert (M, p, iM);
      gsl_blas_dsymv (Uplo, 1.0, iM, a, 1.0, Ma);
      gsl_blas_ddot (a, Ma, &aMa);
      
      gsl_vector_scale (Ma, (1.0/aMa));
      for (int i=0; i<NbIn; i++)
	Filter[i][z] = gsl_vector_get (Ma, i);
    }
  gsl_matrix_free (M);
  gsl_matrix_free (iM);
  gsl_vector_free (a);
  gsl_vector_free (Ma);
  gsl_permutation_free (p);
  return 0; 
}


/////////////////////////////////////////////////
/// compute matched filter : 
/// \param CMat : Input covariance matrices object
/// \param Filter : array of filter coefficients
/// \param MixCol : mixing column
/////////////////////////////////////////////////
template <typename T> Status MF (CovMat<T> CMat, arr2<T> &Filter, vector<double> MixCol, arr<T> &noise, arr2<T> &svd)
{
  int NbIn  = CMat.getNbIn();
  int NbBin = CMat.getNbBin();
  Filter.alloc(NbIn, NbBin);

  gsl_vector * A = gsl_vector_alloc (NbIn);
  gsl_vector * AtinvR = gsl_vector_alloc (NbIn);

  for (int i=0; i<NbIn; i++)
    gsl_vector_set(A, i, MixCol[i]);
  
  gsl_matrix * R = gsl_matrix_alloc (NbIn, NbIn);
  gsl_matrix * V = gsl_matrix_alloc (NbIn, NbIn);
  gsl_vector * S = gsl_vector_alloc (NbIn);
  gsl_vector * work = gsl_vector_alloc (NbIn);

  arr2<T> CArr = CMat.getCovMat();
  double eps = 2.2204e-16;

  for (long z=0; z<NbBin; z++)
    {
      setMatrix(CArr, z, NbIn, R);
      gsl_linalg_SV_decomp (R,  V, S, work);

      double cond;
      int rank = 1;
      double result;
      cond = gsl_vector_get(S,0)/gsl_vector_get(S,NbIn-1);
      for (int i = NbIn - 1; i > 0; i--){
	if (gsl_vector_get(S,i) < (gsl_vector_get(S,0)*eps)) gsl_vector_set(S,i,0.);
	else{rank = i + 1;break;}
      }
      
      gsl_linalg_SV_solve (R, V, S, A, AtinvR);
      gsl_blas_ddot (AtinvR , A, &result);
      
      noise[z]  = result;
      svd[0][z] = cond;
      svd[1][z] = rank;
      
      for (int i=0; i<NbIn; i++)
	Filter[i][z] = gsl_vector_get (AtinvR, i);
    }
  gsl_matrix_free (R);
  gsl_matrix_free (V);
  gsl_vector_free (S);
  gsl_vector_free (work);
  gsl_vector_free (A);
  gsl_vector_free (AtinvR);

  return 0; 
}

/////////////////////////////////////////////////
/// compute spectral mismatch
/// \param CMat  : Input covariance matrices object
/// \param CMat2 : Input covariance matrices object
/// \param wq : number of modes
/////////////////////////////////////////////////
template <typename T> 
double kullback (CovMat<T> CMat, CovMat<T> CMat2, arr<T> wq)
{
  // init mismatch
  double res = 0;
  double tmp = 0;
  int m  = CMat.getNbIn();
  int nq = CMat.getNbBin();
  
  // used to store covariance matrices
  gsl_matrix *hR, *R, *iR;
  hR  = gsl_matrix_alloc (m, m);
  R   = gsl_matrix_alloc (m, m);
  iR  = gsl_matrix_alloc (m, m);

  // used to invert Rmodel
  CBLAS_UPLO_t Uplo=CblasLower;
  int signum;
  gsl_permutation *p;
  p  = gsl_permutation_alloc (m);
  
  // used to compute matrix-matrix product
  CBLAS_SIDE_t Side=CblasLeft; 
    
  const arr2<T> CArr = CMat.getCovMat();
  const arr2<T> CArr2 = CMat2.getCovMat();

 
  // used to compute svd
  gsl_vector_complex *eig;
  gsl_eigen_nonsymm_workspace *work;
  work = gsl_eigen_nonsymm_alloc (m);
  eig  = gsl_vector_complex_alloc (m);

  // compute mismatch
  int q, i, j, k;
  int ncross = (int)m*(m+1)/2;
  for (q=0; q<nq; q++)
    {
    
      tmp = 0;
      
      //init R : 2 cases
      // contaminant covariance is not fixed
      // cmb contribution already contained in smic->rq
      setMatrix(CArr2,  q, m, R);
      setMatrix(CArr,   q, m, hR);

      //invert model
      gsl_linalg_LU_decomp (R, p, &signum);
      gsl_linalg_LU_invert (R, p, iR);
      
      for (i=0; i<m; i++)
	for (j=0; j<=i; j++){
	  gsl_matrix_set(R, i, j, 0) ;
	}
      
      // product
      // now R contains hR*iR
      gsl_blas_dsymm (Side, Uplo, 1, hR, iR, 1, R);
      
      for (i=0; i<m; i++)
	for (j=0; j<=i; j++)
	  gsl_matrix_get(R, i, j);
      
      // svd
      gsl_eigen_nonsymm (R, eig, work);
      
      // sum eigen values
      double ei ;
      for (i=0; i<m; i++){
	ei = gsl_matrix_get(R, i, i);
	tmp += ei - log(ei);}
      tmp -= m;
      tmp *= 0.5*wq[q];

      res += tmp;
    }
  
  gsl_eigen_nonsymm_free (work);
  
  gsl_permutation_free(p);
  gsl_vector_complex_free (eig);
  
  gsl_matrix_free (hR);
  gsl_matrix_free (R);
  gsl_matrix_free (iR);
    
  return res;
  
  
}
