// Copyright (C) 2007,2008 Marc Betoule

// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

// report bugs at betoule@apc.univ-paris7.fr


//Purpose:
//Iteratively fills holes in an healpix map by putting in empty pixels
//the mean value of its neighbors

#define MAX(x,y) x>y?x:y

#include <cstdlib>
#include "fitshandle.h"
#include <vector>
#include <cmath>
#include <sstream>
#include "healpix_map.h"
#include "healpix_map_fitsio.h"

using namespace std;

template <typename T> T fillPixNeig(Healpix_Map<T> & m, vector<int> & l){
  //vector<int>::iterator icenter;
  
  T maxdiff=0;
  int npix = l.size();
  vector<T> val(npix);
#pragma omp parallel
  {
    T lap;
    T localmax = 0;
    fix_arr< int, 8 >neigpix;
    T diff=0;
#pragma omp for
    for(int ipix = 0; ipix < npix; ipix++){
      lap = 0;
      m.neighbors(l[ipix], neigpix);
      for(int i = 0; i < 8; i+=2) lap += m[neigpix[i]];
      lap -= 4.*m[l[ipix]];
      lap /= 5.0;
      //if(0 != mean) diff = abs((mean-m[l[ipix]])/mean);
      diff = abs(lap);
      if (diff>localmax) localmax=diff;
      val[ipix]=lap+m[l[ipix]];
    }
#pragma omp for
    for(int ipix = 0; ipix < npix; ipix++){
      m[l[ipix]] = val[ipix];
    }
#pragma omp critical
    {
      if(localmax>maxdiff) maxdiff = localmax;
    }
  }
  return maxdiff;
}

/*
float fillPixDisc(Healpix_Map<float> & m, vector<int> & l, float radius){
  vector<int>::iterator icenter, ineig;
  vector<int> neigpix;
  float mean;
  float maxdiff=0;
  float diff=0;
    
  for(icenter = l.begin(); icenter < l.end(); icenter++){
    mean = 0;
    m.query_disc(m.pix2ang(*icenter), radius, neigpix);
    for(ineig = neigpix.begin(); ineig < neigpix.end(); ineig++) mean += m[*ineig];
    mean /= (float)(neigpix.size());
    if(mean != 0) diff = abs((mean-m[*icenter])/mean);
    maxdiff = MAX(maxdiff,diff); 
    m[*icenter] = mean;
  }
  return maxdiff;
}
*/

//void solvescale(Healpix_Map<float> & mapin, float eps, float radius, int niter, vector<int> & listEPix){
template <typename T> void solvescale(Healpix_Map<T> &mapin, T eps, int niter, vector<int> & listEPix, bool verbose=0){
  if (verbose)
    cout << "Iter : ";
  //Iterative solve
  T maxdiff = 2*eps;
  for(int iter = 0; (iter != niter) and (maxdiff > eps) ; iter++){
    //if(radius <= 0) maxdiff = fillPixNeig(mapin, listEPix);
    //else maxdiff = fillPixDisc(mapin, listEPix, radius);
    maxdiff = fillPixNeig(mapin, listEPix);
    if (verbose)
      cout << flush << iter << "(" << maxdiff << ")"<< " . ";
  }
  if (verbose)
    cout << "\n";
}

template <typename T> int fastfiller (Healpix_Map<T> &mapin, Healpix_Map<T> &mask,T eps=1e-6, int niter=-1, int startscale=0, bool verbose=0)
{
  //double radius = 0;
  //radius = radius/60*degr2rad; //convert radius in radian

  //store some generic info on the map
  bool swapped = false; // if it was ring ordered first then put it back to ring ordering scheme after computation
  T maxvalue = mapin[0]; // use as a scale for the convergence criterion.
  vector<int> finestListEPix; // list of pixels to fill
  
  if(startscale == 0) startscale = mapin.Nside();
  if(startscale > 256) startscale = 256;
  if(startscale < mapin.Nside()){ //save unusefull computation in case of monoscale resolution
    //translate the map in Nest ordering scheme
    if(mapin.Scheme() == RING){mapin.swap_scheme(); mask.swap_scheme(); swapped=true;}
    //Use Healpix_undef rather than 0
    for(int i = 0; i < mapin.Npix(); i++){
      if(mapin[i] == 0 || approx<T>(mapin[i], Healpix_undef)){ 
	if (mask[i] !=0){
	  mapin[i] = Healpix_undef;
	  finestListEPix.push_back(i);
	}}
      else maxvalue = MAX(maxvalue,mapin[i]);
    }
    
    //Compute the coarsest solution
    Healpix_Map<T> startmap(startscale, mapin.Scheme(), SET_NSIDE);
    startmap.Import_degrade(mapin);
    vector<int> listEPix;
    for(int i = 0; i < startmap.Npix(); i++){
      if(approx<T>(startmap[i], Healpix_undef)){
	startmap[i] = 0;
	listEPix.push_back(i);
      }
    }
    solvescale(startmap, eps*maxvalue, niter, listEPix, verbose);
    //fitshandle outfits;
    //outfits.create("!coarstest.fits");
    //write_Healpix_map_to_fits (outfits,startmap,FITSUTIL<float>::DTYPE);
    //refine this solution
    for(int scale = startscale*2; scale < mapin.Nside(); scale*=2){
      Healpix_Map<T> mapscale(scale, mapin.Scheme(), SET_NSIDE);
      mapscale.Import_degrade(mapin);
      //Build the list of empty pixels and initialize them with the coarser solution
      listEPix.clear();
      for(int i = 0; i < mapscale.Npix(); i++){
	if(approx<T>(mapscale[i], Healpix_undef)){
	  mapscale[i] = startmap[i/4];
	  listEPix.push_back(i);
	}
      }
      //Solve the finer solution
      solvescale(mapscale, eps*maxvalue, niter, listEPix, verbose);
      //ostringstream filename;
      //filename << "!scale" << scale << ".fits";
      //outfits.create(filename.str());
      
      //take this as starting point for the next refinement step
      startmap.swap(mapscale);
      //write_Healpix_map_to_fits (outfits,startmap,FITSUTIL<float>::DTYPE);
    }
    
    //Update the mapin
#pragma omp parallel for
    for(int i = 0; i < mapin.Npix(); i++){
      if(approx<T>(mapin[i],Healpix_undef)){
	mapin[i] = startmap[i/4];
      }
    }
  }
  else{
    for(int i = 0; i < mapin.Npix(); i++){
      if(mapin[i] == 0 || approx<T>(mapin[i], Healpix_undef))
	finestListEPix.push_back(i);
      else 
	maxvalue = MAX(maxvalue,mapin[i]);
    }
  }
  
  //Compute the finest solution
  if (verbose)
    cout << "eps is " << eps*maxvalue << "\n";
  solvescale(mapin, eps*maxvalue, niter, finestListEPix, verbose);

  if(swapped) mapin.swap_scheme();
 
  return 0;
}
