/*
 * Project: spherelib
 * Date: 08/10  
 * Authors: M. Le Jeune
 *
 */

/* Copyright (C) 2008 APC CNRS Université Paris Diderot <lejeune@apc.univ-paris7.fr>  
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see http://www.gnu.org/licenses/gpl.html
 */

#include <string>
#include <sstream>
#include <fitshandle.h>
#include <iostream>
#include <iterator>
#include "healpix_map.h"
#include "xcomplex.h"
#include "alm.h" 
#include "alm_healpix_tools.h"
#include "alm_powspec_tools.h"
#include "powspec.h"
#include "healpix_base.h"
#include "healpix_map_fitsio.h"
#include "alm_fitsio.h"
#include "levels_facilities.h"
#include "lsconstants.h"
#include "powspec_fitsio.h"
#include "CD_CommonDefinition.h"
#include "AlmExt.h"
#include "MapExt.h"
#include "PowSpecExt.h"
#include "Bin.h"
#include "CovMat.h"
#include "mask.h"
#include "cmatrix.h"

#include <gsl/gsl_rng.h>
#include <gsl/gsl_sort_double.h>


using namespace std;

void print_time (time_t start, time_t startu, time_t end, time_t endu){
  double elapsed_seconds  = end-start; 
  double elapsed_useconds = endu-startu;
  fprintf(stdout, "elapsed time %lf s\n", elapsed_seconds + elapsed_useconds/1000000.0);
}
void rm_tmp_file (string str){
  char tplt[100];
  sprintf(tplt, "rm -f %s", str.c_str());
  system(tplt);
}

void rm_lst_tmp (int nfile, string** arr){
  char tplt[100];
  for (int i=0; i<nfile; i++) {
    sprintf(tplt, "rm -r %s", (*arr)[i].c_str());
    system(tplt);
  }
}


string get_tmp_file (string str){
  char* tempdir = getenv("TMPDIR");
  if (tempdir==NULL)
    tempdir = "/tmp";
  char tplt[100];
  sprintf(tplt, "%s/%s%d", tempdir, str.c_str(), rand());
  string out = tplt;
  return out;
}

void get_lst_tmp (int nfile, string str, string** arr){
  (*arr)    = new string[nfile]; 
  char* tempdir = getenv("TMPDIR");
  if (tempdir==NULL)
    tempdir = "/tmp";
  char tplt[100];
  for (int i=0; i<nfile; i++) {
    sprintf(tplt, "%s/%s%d", tempdir, str.c_str(), rand());
    (*arr)[i] = tplt;
  }
}


// parse the arguments of the parameter file which are a list of strings (blank separated) 
// (or any kind of list that could be convert from string afterwards)
int parse_list (string str, string** arr){
  // Get size of list
  int    i   = 0;
  int   pos  = 0;
  while (pos < (int)str.size())    {
    pos = str.find(" ", pos); i++; // pos = -1 if not found
    if (pos>0) 
      pos++;
    else
      pos = (int)str.size()+1;    // endwhile
  }
  // Knowing the size of the list...
  int nbsub = i;
  (*arr)    = new string[nbsub]; 
  for (i=0; i<nbsub; i++)    {
    string::size_type eqpos=str.find(" ");
    if (eqpos!=string::npos)      {
      (*arr)[i]= trim(str.substr(0,eqpos));
      str      = (str.substr(eqpos+1,string::npos));    }
    else
      (*arr)[i] = trim(str); // last element
  }
  return nbsub;
}



template <typename T> int readfitscol (string file, arr<T> &data, int colnum=1)
{
  fitshandle inp;
  inp.open(file);
  inp.goto_hdu(2);
  if (inp.key_present("PDMTYPE"))
    colnum++;
  inp.read_entire_column(colnum,data);
  return 0;
}

template <typename T> int writefitscol (string file, arr<T> &data, int colnum=1,string title="", string unit="")
{
  fitshandle out;
  out.create(file);
  vector<fitscolumn> cols;
  cols.push_back(fitscolumn(title,unit,1,planckType<T>()));
  out.insert_bintab(cols);
  out.write_column(colnum,data);
  return 0;
}

template <typename T> int writefitstable (string file, arr2<T> &data, int colnum=1,string title="",string unit="")
{
  fitshandle out;
  out.create(file);
  vector<fitscolumn> cols;
  std::stringstream oti;
  for (int col=0; col<data.size1(); col++)
    {
      oti.str("");
      oti << title << col;
      cols.push_back(fitscolumn (oti.str(),unit,1,planckType<T>()));
    }
  out.insert_bintab(cols);
  for (int col=0; col<data.size1(); col++)
    out.write_column_raw(colnum++,data[col],data.size2());
  return 0;
}

template<typename T> void readfitstable(string file, arr2<T> &data, int colnum=1)
{
  fitshandle inp;
  inp.open(file);
  inp.goto_hdu(2);
  int ncols = inp.ncols()-colnum+1;
  int nrows = inp.nrows();
  arr<T> datacol;
  data.alloc(ncols, nrows); 
  for (long j=0; j<ncols; j++){
    inp.read_entire_column(j+colnum,datacol);
    for (long i=0; i<nrows; i++)   
      data[j][i] = datacol[i];
  } 
}




/////////////////////////////////////////////////
/// Compute the ILC filter.
/// \param infile : covariance matrices files
/// \param nbin   : number of entries
/// \param outfile : out filter file
/// \param mixcol : emission law
template <typename T> int stat2ILCfilter (string infile, long nbin, string outfile, arr<double> mixcol)
{
  arr2<T> covmat;
  CovMat<T> stats;
  readfitstable (infile, covmat, 2);
  int nZ   = covmat.size2();	 
  
  stats.setCovMat(covmat);
  vector<double> vmix(nbin);
  for (int i=0; i<nbin; i++)
    vmix[i] = mixcol[i];


  arr2<T> filter (nbin, nZ);
  ILC(stats, filter, vmix);
  writefitstable(outfile, filter);
}


/////////////////////////////////////////////////
/// Compute the ILC filter.
/// \param infile : covariance matrices files
/// \param nbin   : number of entries
/// \param outfile : out filter file
/// \param mixcol : emission law
template <typename T> T mismatch (string infile, string infile2)
{
  arr2<T> covmat;
  CovMat<T> stats;
  CovMat<T> model;
  readfitstable (infile , covmat, 2);
  stats.setCovMat(covmat);
  int nZ   = covmat.size2();	 

  readfitstable (infile2, covmat, 2);
  model.setCovMat(covmat);
  
  arr<T> wq;
  readfitscol (infile, wq);

  T m = 0;
  m = kullback (stats, model, wq);
  return m;
}

/////////////////////////////////////////////////
/// Compute the ILC filter.
/// \param infile : covariance matrices files
/// \param nbin   : number of entries
/// \param outfile : out filter file
/// \param mixcol : weighted emission law
/// \param noisefile : out noise file
/// \param svdfile : out svd file (conditionning + rank)
template <typename T> int stat2MFfilter (string infile, long nbin, string outfile, double* mixcol, string noisefile, string svdfile)
{
  arr2<T> covmat;
  CovMat<T> stats;
  readfitstable (infile, covmat, 2);
  int nZ   = covmat.size2();	 
  
  stats.setCovMat(covmat);
  vector<double> vmix(nbin);
  for (int i=0; i<nbin; i++)
    vmix[i] = mixcol[i];
  arr2<T> filter (nbin, nZ);
  arr<T>  noise (nZ);
  arr2<T> svd(2, nZ); 
  MF(stats, filter, vmix, noise, svd);
  writefitstable(outfile, filter);
  writefitscol(noisefile,  noise);
  writefitstable(svdfile, svd);
}


/////////////////////////////////////////////////
/// Compute the linear combination of alm coefficients. 
/// \param infile : list of alm file names
/// \param nbin   : number of input files
/// \param outfile : out alm file name
/// \param mixcol : list of coefficients
/// \param polar  : polarization option
template <typename T> int combinealm (string* infile, long nbin, string outfile, arr<T> mixcol, bool polar)
{
  fitshandle out;
  out.create (outfile);
  AlmExt<T> alm;

  int lmax;
  int biin = 0;
  int tmplmax, mmax;
  get_almsize(infile[0], lmax, mmax );  
  for (int iin=1; iin < nbin; iin++) {
    get_almsize(infile[iin], tmplmax, mmax );  
    if (tmplmax>lmax)
      {
	lmax = tmplmax;
	biin = iin;
      }
  }
  get_almsize(infile[biin], lmax, mmax );  
  
  if (polar)   {
    AlmExt<T> almST, almSE, almSB;
    if (nbin>0){
      read_Alm_from_fits (infile[biin], almST, lmax, mmax, 2);
      read_Alm_from_fits (infile[biin], almSE, lmax, mmax, 3);
      read_Alm_from_fits (infile[biin], almSB, lmax, mmax, 4);
      
      almST*= mixcol[biin];
      almSE*= mixcol[biin];
      almSB*= mixcol[biin];
      if (nbin>1) {
	for (int iin=0; iin < nbin; iin++) {
	  if (iin != biin) {
	    read_Alm_from_fits (infile[iin], alm, lmax, mmax, 2);
	    almST.LinComb(alm, mixcol[iin]);
	    read_Alm_from_fits (infile[iin], alm, lmax, mmax, 3);
	    almSE.LinComb(alm, mixcol[iin]);
	    read_Alm_from_fits (infile[iin], alm, lmax, mmax, 4);
	    almSB.LinComb(alm, mixcol[iin]);}   } }

      write_Alm_to_fits (out,almST,almST.Lmax(),almST.Mmax(),planckType<T>());
      write_Alm_to_fits (out,almSE,almSE.Lmax(),almSE.Mmax(),planckType<T>());
      write_Alm_to_fits (out,almSB,almSB.Lmax(),almSB.Mmax(),planckType<T>());
    }}
  else { 
    AlmExt<T> almS;
    if (nbin>0) {
      read_Alm_from_fits (infile[biin], almS, lmax, mmax, 2);
      almS*= mixcol[biin];
      if (nbin>1) {
	for (int iin=0; iin < nbin; iin++) {
	  if (iin != biin) {
	    read_Alm_from_fits (infile[iin], alm, lmax, mmax, 2);
	    almS.LinComb(alm, mixcol[iin]);}}}
      write_Alm_to_fits (out,almS,almS.Lmax(),almS.Mmax(),planckType<T>());}}
  return 0;
}

/////////////////////////////////////////////////
/// Perform linear combination over pixel domain zones
/// \param infile : list of input Healpix maps
/// \param nbin : number of input maps
/// \param maskfile : Healpix map which defined spatial zones
/// \param mixfile : FITS file containing coefficients
/// \param outfile : out file name
template <typename T> int combinezone (string* infile, int nbin, string outfile,string mixfile, string maskfile)
{
  // Load zone file
  MapExt<int> Mask;
  read_Healpix_map_from_fits(maskfile, Mask, 1, 2);

  // get number of zones
  int min , max;
  Mask.minmax(min,max);
  long nzone = (long) max+1;

  arr<T>  mixcol(nzone);
  MapExt<T> mapS ;
  mapS.SetNside(Mask.Nside(), Mask.Scheme());
  mapS.fill(0);
  MapExt<T> map;
  if (nbin>0) {
    for (int iin=0; iin < nbin; iin++) {
      readfitscol (mixfile, mixcol, iin+1);
      read_Healpix_map_from_fits(infile[iin], map, 1, 2);
      mapS.LinCombZone(map, &mixcol[0], Mask);}
  }
  fitshandle out;
  out.create(outfile);
  write_Healpix_map_to_fits (out, mapS,planckType<T>());
}

/////////////////////////////////////////////////
/// Perform linear combination over pixel domain zones
/// \param infile : list of input Healpix maps
/// \param nbin : number of input maps
/// \param nside : resolution of the out map
/// \param mixfile : FITS file containing coefficients
/// \param outfile : out file name
template <typename T> int combinezone (string* infile, int nbin, string outfile,string mixfile)
{
  arr<T>  mixcol;
  MapExt<T> mapS ;
  MapExt<T> map;
  AlmExt<T> alm;
  if (nbin>0) {
    read_Healpix_map_from_fits(infile[0], mapS, 1, 2);
    mapS.fill(0);
    for (int iin=0; iin < nbin; iin++) {
      readfitscol (mixfile, mixcol, iin+1);
      MapExt<T> W (mixcol, RING);
      int lmax = W.Nside()*2;
      alm.FromMap(W, lmax, lmax);
      W.FromAlm(alm, mapS.Nside());
      read_Healpix_map_from_fits(infile[iin], map, 1, 2);
      map *= W;
      mapS.LinComb (map, 1);}
  }
  fitshandle out;
  out.create(outfile);
  write_Healpix_map_to_fits (out, mapS,planckType<T>());
}

/////////////////////////////////////////////////
/// Perform linear combination over multipole domain
/// \param infile : list of input spherical harmonic coefficients
/// \param nbin : number of input maps
/// \param mixfile : FITS file containing coefficients
/// \param outfile : out file name
template <typename T> int combinemultipole (string* infile, int nbin, string outfile,string mixfile)
{
  // get number of zones
  int lmax , mmax;
  get_almsize(infile[0], lmax, mmax );  
  AlmExt<T>  Alm[nbin];
  arr2<T> Filter (nbin, lmax+1);
  readfitstable(mixfile, Filter);
  
  for (int i=0; i<nbin; i++)
    read_Alm_from_fits (infile[i], Alm[i], lmax, mmax, 2);
  
  AlmExt<T> almS (lmax, mmax);
  almS.multiScaleL (Alm, Filter, nbin);

  fitshandle out;
  out.create(outfile);
  write_Alm_to_fits (out,almS,almS.Lmax(),almS.Mmax(),planckType<T>());
}


/////////////////////////////////////////////////
/// Compute 2nd order statistics
/// \param infile : list of input Healpix maps
/// \param nbin : number of input maps
/// \param outfile : out file name
/// \param maskfile : Healpix map which defined spatial zones
template <typename T> int map2stat (string* infile, int nbin, string outfile,string maskfile)
{
  MapExt<T> Maps[nbin];
  for (int i=0; i<nbin; i++)
    read_Healpix_map_from_fits(infile[i], Maps[i], 1, 2);
  MapExt<int> Mask;
  read_Healpix_map_from_fits(maskfile, Mask, 1, 2);
  
  CovMat<T> CovMat;
  CovMat.FromMap (Maps, nbin, Mask);
  arr2<T> Mat = CovMat.getCovMat();
  arr<T>  W   = CovMat.getNbMode();

  fitshandle out;
  out.create(outfile);
  vector<fitscolumn> cols;
  cols.push_back (fitscolumn ("Number of mode", "no unit", 1, planckType<T>())); 
  std::stringstream oti;
  for (int col=0; col<Mat.size1(); col++)
    {
      oti.str("");
      oti << "Cross term " << col;
      cols.push_back(fitscolumn (oti.str(),"no unit",1,planckType<T>()));
    }
  
  out.insert_bintab(cols);
  out.write_column_raw (1, &W[0], W.size());
  for (int col=0; col<Mat.size1(); col++)
    out.write_column_raw(col+2,Mat[col],Mat.size2());
  out.close();
  return 0;
  
}


/////////////////////////////////////////////////
/// Compute 2nd order statistics
/// \param infile : list of input Healpix maps
/// \param nbin : number of input maps
/// \param outfile : out file name
/// \param fwhm : fwhm of the Gaussian window function 
/// \param lmax : maximum multipole order for analysis
template <typename T> int map2statgauss (string* infile, int nbin, string outfile,double fwhm, int lmax,int nside)
{
  MapExt<T> Maps[nbin];
  for (int i=0; i<nbin; i++)
    read_Healpix_map_from_fits(infile[i], Maps[i], 1, 2);

  CovMat<T> CovMat;
  CovMat.FromMap (Maps, nbin, fwhm, lmax, nside);

  arr2<T> Mat = CovMat.getCovMat();
  arr<T>  W   = CovMat.getNbMode();

  fitshandle out;
  out.create(outfile);
  vector<fitscolumn> cols;
  cols.push_back (fitscolumn ("Number of mode", "no unit", 1, planckType<T>())); 
  std::stringstream oti;
  for (int col=0; col<Mat.size1(); col++)
    {
      oti.str("");
      oti << "Cross term " << col;
      cols.push_back(fitscolumn (oti.str(),"no unit",1,planckType<T>()));
    }
  
  out.insert_bintab(cols);
  out.write_column_raw (1, &W[0], W.size());
  for (int col=0; col<Mat.size1(); col++)
    out.write_column_raw(col+2,Mat[col],Mat.size2());
  out.close();
  return 0;
  
}


int lmax2nside (int lmax)
{
  int Nside;
  int n = (int) ceil(log2( lmax/2.0 ));
  Nside = (int) pow((double) 2,(double) n) ;
  return Nside;
}

/////////////////////////////////////////////////
/// Compute 2nd order statistics
/// \param infile : list of input Healpix variance maps
/// \param nbin : number of input maps
/// \param outfile : out file name
/// \param maskfile : Healpix map which defined spatial zones
/// \param filterfile : needlet filter coefficients
template <typename T> int varmap2stat (string* infile, int nbin, string outfile,string maskfile,string filterfile)
{

  MapExt<int> Mask;
  read_Healpix_map_from_fits(maskfile, Mask, 1, 2);
  int min, max;
  Mask.minmax(min,max);
  long nzone = (long) max+1;

  // init filter reading specfile
  arr<T> filter;
  readfitscol(filterfile, filter, 1); 

  // compute square filter
  int lmax = filter.size();
  // replace with last non nul value
  for (int ell=lmax; ell>=0 ; ell--)
    if (filter[ell] > 0.0)
      {
	lmax = ell;
	break;
      }
  AlmExt<T> Alm(lmax, lmax);
  Alm.SetToZero();
  for (int ell=0; ell<=lmax; ell++)
    Alm(ell,0) = filter[ell]*sqrt(((2*ell)+1)/(4*pi));
  MapExt<T> Map;
  Map.FromAlm(Alm,lmax2nside(2*lmax));
  Map.Prod(Map);
  
  Alm.FromMap(Map, 2*lmax, 2*lmax);
  arr<T> squarefilter(2*lmax+1);
  for (int ell=0; ell<=2*lmax; ell++){
    T f = sqrt(norm(Alm(ell,0)));
    squarefilter[ell] = f/sqrt(2*ell+1)*sqrt(4*pi);}
  
  // filter map
  T cst ;
  arr<T>  meanzone(nzone);
  arr2<T> MeanZone(nbin, nzone);
  for (int i=0; i<nbin; i++){
    read_Healpix_map_from_fits(infile[i], Map, 1, 2);
    cst = 4*pi/(12*Map.Nside()*Map.Nside());
    Alm.FromMap (Map, 2*lmax, 2*lmax); // get alm
    Alm.ScaleL (squarefilter); // compute filtered alm
    Map.FromAlm (Alm, Mask.Nside());// compute Healpix map
    Map.MeanZone(meanzone, Mask);
    
    for (int z=0; z<nzone; z++)
      MeanZone[i][z] = cst*meanzone[z];
  }
  writefitstable (outfile, MeanZone, 1,"Mean " ,"no unit");
  return 0;
}


/////////////////////////////////////////////////
/// Compute 2nd order statistics
/// \param infile : list of input spherical harmonic coefficients
/// \param nbin : number of input maps
/// \param outfile : out file name
template <typename T> int alm2stat (string* infile, int nbin, string outfile)
{
  AlmExt<T> Alm[nbin];
  int lmax, mmax;
  get_almsize(infile[0], lmax, mmax );  
  for (int i=0; i<nbin; i++)
    read_Alm_from_fits (infile[i], Alm[i], lmax, mmax, 2);
  
  CovMat<T> CovMat;
  CovMat.FromAlm (Alm, nbin);

  arr2<T> Mat = CovMat.getCovMat();
  arr<T>  W   = CovMat.getNbMode();

  fitshandle out;
  out.create(outfile);
  vector<fitscolumn> cols;
  cols.push_back (fitscolumn ("Number of mode", "no unit", 1, planckType<T>())); 
  std::stringstream oti;
  for (int col=0; col<Mat.size1(); col++)
    {
      oti.str("");
      oti << "Cross term " << col;
      cols.push_back(fitscolumn (oti.str(),"no unit",1,planckType<T>()));
    }
  
  out.insert_bintab(cols);
  out.write_column_raw (1, &W[0], W.size());
  for (int col=0; col<Mat.size1(); col++)
    out.write_column_raw(col+2,Mat[col],Mat.size2());
  out.close();
  return 0;
  
}



/////////////////////////////////////////////////
/// Generate spline window functions from ell bands
/// \param bands : array of ell bands values
/// \param files : out file name
/// \param onefile : save filters in one file (onefile=1), or one window function per file (onefile=0)
int makesplinefilters (arr<int> &bands, string* files, bool onefile=1)
{
  Bin BO;
  arr2<double> bins ;
  Spline(bands, bins);
  BO.Set(bins); // generate spline window functions
  int J   = BO.getNBand ();
  int res = 0;
  // save result in one file
  if (onefile)
    {
      arr2<double> TFilter = BO.getBins();
      writefitstable (files[0], TFilter, 1,"Window function number " ,"no unit");
    }
  // save in different files
  else
    {
      arr<double>* Filter=NULL;
      for (int j=0; j<J; j++)
	{
	  Filter = BO.getQFilter(j);
	  if (!onefile)
	    res    = writefitscol (files[j], *Filter);
	  delete Filter;                            // free memory
	  Filter = NULL;
	} 
    }
  return res;
}

/////////////////////////////////////////////////
/// Generate prolat window functions from ell bands
/// \param bands : array of ell bands values
/// \param files : out file name
/// \param onefile : save filters in one file (onefile=1), or one window function per file (onefile=0)
int makeprolatefilters (arr<int> bands, string* files, bool onefile, arr<float> thetas, int rule)
{
  Bin BO;
  arr2<double> bins ;
  Prolate(bands, bins, thetas, rule);
  BO.Set(bins); // generate spline window functions
  int J   = BO.getNBand ();
  int res = 0;
  // save result in one file
  if (onefile)
    {
      arr2<double> TFilter = BO.getBins();
      writefitstable (files[0], TFilter, 1,"Window function number " ,"no unit");
    }
  // save in different files
  else
    {
      arr<double>* Filter=NULL;
      for (int j=0; j<J; j++)
	{
	  Filter = BO.getQFilter(j);
	  if (!onefile)
	    res    = writefitscol (files[j], *Filter);
	  delete Filter;                            // free memory
	  Filter = NULL;
	} 
    }
  return res;
}

/////////////////////////////////////////////////
/// Filter each map in harmonic space using spectra from file
/// \param infile : list of map file names
/// \param outfile : out map file name
/// \param filterfile : list of files which will be used as filter coefficients
/// \param lmax : maximum order of analysis
/// \param nside : Healpix map resolution parameter
/// \param polar : polarization option
/// \param niter : iteration number
template <typename T> int filtermap (string infile, string outfile, string filterfile,int lmax,int nside,int polar, int niter=0)
{

  fitshandle out;

  // init filter reading specfile
  arr<T> filter (lmax);
  readfitscol(filterfile, filter, 1); 

  if (polar)
    {
      AlmExt<T> almT, almE, almB;
      MapExt<T> mapI, mapQ, mapU;
      read_Healpix_map_from_fits(infile, mapI, 1, 2);
      read_Healpix_map_from_fits(infile, mapQ, 2, 2);
      read_Healpix_map_from_fits(infile, mapU, 3, 2);

      almT.FromPolarMap (mapI, mapQ, mapU, almE, almB, lmax, lmax, niter);
      
      almT.ScaleL (filter);
      almE.ScaleL (filter);
      almB.ScaleL (filter);

      mapI.FromPolarAlm (almT, almE, almB, mapQ, mapU, nside);
    
      out.create (outfile);
      write_Healpix_map_to_fits (out, mapI,mapQ,mapU,planckType<T>()); 

    }
  else
    {
      AlmExt <T> alm;     // filtered alm
      MapExt<T>  map;   
      read_Healpix_map_from_fits(infile, map, 1, 2);
      
      alm.FromMap (map, lmax, lmax, niter); // get alm
      
      alm.ScaleL (filter); // compute filtered alm
      map.FromAlm (alm, nside);// compute Healpix map
      out.create (outfile);
      write_Healpix_map_to_fits (out, map,planckType<T>());// save it to file
      
    }
  return(0);
  
}

/////////////////////////////////////////////////
/// Upgrade or downgrade Healpix map to new resolution
/// \param infile : map file name
/// \param outfile : out map file name
/// \param pessimistic : set to healpix undef when downgrading
/// \param nside : Healpix map resolution parameter
/// \param polar : polarization option
template <typename T> int udgrade (string infile, string outfile, bool pessimistic,int nside,int polar)
{

  fitshandle out;
  out.create (outfile);

  if (polar)
    {
      MapExt<T> mapI;
      Healpix_Map<T> outmapI, outmapQ, outmapU;
      
      read_Healpix_map_from_fits(infile, mapI, 1, 2);

      outmapI.SetNside(nside, mapI.Scheme());
      outmapI.Import(mapI,pessimistic);

      read_Healpix_map_from_fits(infile, mapI, 2, 2);
      outmapQ.SetNside(nside, mapI.Scheme());
      outmapQ.Import(mapI,pessimistic);

      read_Healpix_map_from_fits(infile, mapI, 3, 2);
      outmapU.SetNside(nside, mapI.Scheme());
      outmapU.Import(mapI,pessimistic);
      
      write_Healpix_map_to_fits (out, outmapI,outmapQ,outmapU,planckType<T>()); 

    }
  else
    {
      MapExt<T>  map;   
      MapExt<T> outmap;
      read_Healpix_map_from_fits(infile, map, 1, 2);
      outmap.SetNside(nside, map.Scheme());
      outmap.Import(map,pessimistic);
      write_Healpix_map_to_fits (out, map,planckType<T>());// save it to file
      
    }
  return(0);
  
}


/////////////////////////////////////////////////
/// Compute the term by term product between Healpix maps. 
/// \param infile : list of map file names
/// \param nbin   : number of input files
/// \param outfile : out map file name
/// \param polar  : polarization option
template <typename T> int prodmap (string* infile, long nbin, string outfile, bool polar)
{

  fitshandle out;
  MapExt<T> map;

  if (polar)   {
    MapExt<T> mapSI, mapSQ, mapSU;
    read_Healpix_map_from_fits(infile[0], mapSI, 1, 2);
    read_Healpix_map_from_fits(infile[0], mapSQ, 2, 2);
    read_Healpix_map_from_fits(infile[0], mapSU, 3, 2);
    if (nbin>1){
      for (int iin=1; iin < nbin; iin++) {
	read_Healpix_map_from_fits(infile[iin], map, 1, 2);
	mapSI*=map;
	read_Healpix_map_from_fits(infile[iin], map, 2, 2);
	mapSQ*=map;
	read_Healpix_map_from_fits(infile[iin], map, 3, 2);
	mapSU*=map;}}
    out.create (outfile);
    write_Healpix_map_to_fits (out, mapSI,mapSQ,mapSU,planckType<T>()); }
  else { 
    MapExt<T> mapS;
    read_Healpix_map_from_fits(infile[0], mapS, 1, 2);
    if (nbin>1) {
      for (int iin=1; iin < nbin; iin++) {
	read_Healpix_map_from_fits(infile[iin], map, 1, 2);
	mapS*=map;}
    }
    out.create (outfile);
    write_Healpix_map_to_fits (out, mapS,planckType<T>());}
  return 0;
  
}


/////////////////////////////////////////////////
/// Compute the linear combination of Healpix maps. 
/// \param infile : list of map file names
/// \param nbin   : number of input files
/// \param outfile : out map file name
/// \param mixcol : list of coefficients
/// \param polar  : polarization option
template <typename T> int combinemap (string* infile, long nbin, string outfile, double* mixcol, bool polar)
{
  fitshandle out;
  out.create (outfile);
  MapExt<T> map;

  if (polar)   {
    MapExt<T> mapSI, mapSQ, mapSU;
    if (nbin>0){
      read_Healpix_map_from_fits(infile[0], mapSI, 1, 2);
      read_Healpix_map_from_fits(infile[0], mapSQ, 2, 2);
      read_Healpix_map_from_fits(infile[0], mapSU, 3, 2);
      mapSI*= mixcol[0];
      mapSU*= mixcol[0];
      mapSQ*= mixcol[0];
      if (nbin>1) {
	for (int iin=1; iin < nbin; iin++) {
	  read_Healpix_map_from_fits(infile[iin], map, 1, 2);
	  mapSI.LinComb(map, mixcol[iin]);
	  read_Healpix_map_from_fits(infile[iin], map, 2, 2);
	  mapSQ.LinComb(map, mixcol[iin]);
	  read_Healpix_map_from_fits(infile[iin], map, 3, 2);
	  mapSU.LinComb(map, mixcol[iin]);}	}
      write_Healpix_map_to_fits (out, mapSI,mapSQ,mapSU,planckType<T>()); }}

  else { 
    MapExt<T> mapS;
    if (nbin>0) {
      read_Healpix_map_from_fits(infile[0], mapS, 1, 2);
      mapS*= mixcol[0];
      if (nbin>1) {
	for (int iin=1; iin < nbin; iin++) {
	  read_Healpix_map_from_fits(infile[iin], map, 1, 2);
	  mapS.LinComb(map, mixcol[iin]);}}
      write_Healpix_map_to_fits (out, mapS,planckType<T>());}}
  return 0;
}

/////////////////////////////////////////////////
/// Remove multipole of HEALPix masked map up to order
/// \param infile : map file name
/// \param outfile : out map file name
/// \param maskfile : mask file name
/// \param order : order
template <typename T> int regressmap (string infile, string outfile, string maskfile, int order)
{
  
  fitshandle out;
  MapExt<T> map;
  read_Healpix_map_from_fits(infile, map, 1, 2);
  MapExt<T> mask;
  read_Healpix_map_from_fits(maskfile, mask, 1, 2);
  Regress (map, mask, order);
  out.create (outfile);
  write_Healpix_map_to_fits (out, map,planckType<T>());
  return 0;
}


/////////////////////////////////////////////////
/// Compute Healpix map from spherical harmonic coefficients 
/// \param infile : alm file name
/// \param outfile : out map file name
/// \param nside : Healpix map resolution parameter
/// \param polar  : polarization option
template <typename T> int alm2map (string infile, string outfile, int nside, bool polar)
{

  int lmax, mmax;
  get_almsize(infile, lmax, mmax );
  
  fitshandle out;
  out.create(outfile);

  if (polar) {
    MapExt <T> mapI, mapQ, mapU;
    AlmExt <T> almT, almE, almB;
    read_Alm_from_fits (infile, almT, lmax, mmax, 2);
    read_Alm_from_fits (infile, almE, lmax, mmax, 3);
    read_Alm_from_fits (infile, almB, lmax, mmax, 4);
    
    mapI.FromPolarAlm(almT, almE, almB, mapQ, mapU, nside); // compute Healpix maps
    write_Healpix_map_to_fits (out, mapI,mapQ,mapU,planckType<T>());
  }  
  else    {
    MapExt <T> map; 
    AlmExt <T> alm;
    read_Alm_from_fits (infile, alm, lmax, mmax);
    map.FromAlm (alm, nside);// compute Healpix map
    write_Healpix_map_to_fits (out, map,planckType<T>());// save it to file
  } 
  return 0;
}

/////////////////////////////////////////////////
/// Compute power spectra from spherical harmonic coefficients 
/// \param infile : alm file name
/// \param outfile : out cl file name
/// \param polar  : polarization option
template <typename T> int alm2cl (string infile, string outfile, bool polar)
{

  int lmax, mmax;
  get_almsize(infile, lmax, mmax );
  
  fitshandle out;
  out.create(outfile);

  if (polar) {

    AlmExt <T> almT, almE, almB;
    read_Alm_from_fits (infile, almT, lmax, mmax, 2);
    read_Alm_from_fits (infile, almE, lmax, mmax, 3);
    read_Alm_from_fits (infile, almB, lmax, mmax, 4);

    PowSpecExt<T> spec;
    spec.FromPolarAlm(almT, almE, almB); // compute spectra if asked
    write_powspec_to_fits (out,spec,4);}// save them to file      
  
  else    {
    AlmExt <T> alm;
    read_Alm_from_fits (infile, alm, lmax, mmax);
    PowSpecExt<T> spec;
    spec.FromAlm(alm);// compute spectrum if asked
    write_powspec_to_fits (out,spec,1);}// save it to file  
  
  
  return 0;
}



/////////////////////////////////////////////////
/// Deconvolve alm using beam from file up to threshold
/// \param infile : alm file name
/// \param outfile : out alm file name
/// \param beamfile : beam file name
/// \param refbeamfile : new beam to apply to alm (can be empty)
/// \param threshold : maximum inverse filter value
template <typename T> int unbeamalm (string infile, string outfile, string beamfile, string refbeamfile, float threshold)
{
  AlmExt<T> fialm ;

  int lmax, mmax;
  get_almsize(infile, lmax, mmax );
  read_Alm_from_fits (infile, fialm, lmax, mmax);
  
  // init filter reading specfile
  arr<T> beam    (lmax);
  arr<T> refbeam (lmax);
  readfitscol(beamfile, beam, 1); 
  if (refbeamfile.size()>1)
    readfitscol(refbeamfile, refbeam, 1); 
  else
    refbeam.fill(1);
  fialm.Deconvol(beam, refbeam, threshold);// unbeam and apply a new beam to alm
  
  fitshandle out;
  out.create (outfile);// save them to file
  write_Alm_to_fits(out, fialm, lmax, mmax, planckType<T>());
  
  return(0);
}



/////////////////////////////////////////////////
/// Filter each alm using spectra from file
/// \param infile : alm file name
/// \param outfile : out alm file name
/// \param filterfile : fits file containing filter coeff in first column
template <typename T> int filteralm (string infile, string outfile, string filterfile, bool polar)
{
  AlmExt <T> fialm;     // filtered alm
  int lmax, mmax;
  
  get_almsize(infile, lmax, mmax );

  // init filter reading specfile
  arr<T> filter (lmax);
  readfitscol(filterfile, filter, 1); 

  if (polar)
    { 
      AlmExt <T> almT, almE, almB;
      read_Alm_from_fits (infile, almT, lmax, mmax, 2);
      read_Alm_from_fits (infile, almE, lmax, mmax, 3);
      read_Alm_from_fits (infile, almB, lmax, mmax, 4);
      
      // compute filtered alm
      almT.ScaleL (filter);
      almE.ScaleL (filter);
      almB.ScaleL (filter);
      
      fitshandle out;
      out.create (outfile);// save them to file
      write_Alm_to_fits (out,almT,lmax,mmax,planckType<T>());
      write_Alm_to_fits (out,almE,lmax,mmax,planckType<T>());
      write_Alm_to_fits (out,almB,lmax,mmax,planckType<T>());
      
    }
  else
    {
      read_Alm_from_fits (infile, fialm, lmax, mmax);
      
      // compute filtered alm
      fialm.ScaleL (filter);
      fitshandle out;
      out.create (outfile);// save them to file
      write_Alm_to_fits(out, fialm, lmax, mmax, planckType<T>());
    }
  return(0);
}



/////////////////////////////////////////////////
/// Compute spherical harmonic coefficients from Healpix map
/// \param infile : map file name
/// \param outfile : out alm file name
/// \param lmax : maximum order of analysis
/// \param niter : iteration number
/// \param clfile : spectrum file name (can be empty)
/// \param polar  : polarization option
template <typename T> int map2alm (string infile, string outfile, int lmax, int niter,string clfile,bool polar)
{
  fitshandle out;
  
  // polarized case
  if (polar)
    {
      AlmExt <T> almT, almE, almB;
      MapExt <T> mapI, mapQ, mapU;
      read_Healpix_map_from_fits(infile, mapI, 1, 2);
      read_Healpix_map_from_fits(infile, mapQ, 2, 2);
      read_Healpix_map_from_fits(infile, mapU, 3, 2);

      almT.FromPolarMap (mapI,mapQ,mapU,almE,almB, lmax, lmax, niter); // get alm
      
      out.create (outfile);// save them to file
      write_Alm_to_fits (out,almT,almT.Lmax(),almT.Mmax(),planckType<T>());
      write_Alm_to_fits (out,almE,almE.Lmax(),almE.Mmax(),planckType<T>());
      write_Alm_to_fits (out,almB,almB.Lmax(),almB.Mmax(),planckType<T>());
      
      PowSpecExt<T> spec;
      if (clfile.size()>1)      {
	spec.FromPolarAlm(almT, almE, almB); // compute spectra if asked
	out.create (clfile);
	write_powspec_to_fits (out,spec,4);}// save them to file      
    }
  else
    {
    
      AlmExt <T> alm; 
      MapExt <T> map;
      read_Healpix_map_from_fits(infile, map, 1, 2);
      alm.FromMap (map, lmax, lmax, niter); // get alm
      out.create  (outfile);// save them to file
      write_Alm_to_fits (out,alm,alm.Lmax(),alm.Mmax(),planckType<T>());
      PowSpecExt<T> spec;
      if (clfile.size()>1){
	spec.FromAlm(alm);// compute spectrum if asked
	out.create (clfile);
	write_powspec_to_fits (out,spec,1);}// save it to file  
    }
  return 0;
  
}

/////////////////////////////////////////////////
/// Compute spherical harmonic coefficients from Healpix map
/// \param infile : map file name
/// \param outfile : out alm file name
/// \param lmax : maximum order of analysis
/// \param nside : needlet resolution
/// \param niter : iteration number
/// \param filterfile : needlet window function
template <typename T> int map2need (string infile, string outfile, int lmax, int nside, int niter,string filterfile)
{
  fitshandle out;
  
  AlmExt <T> alm; 
  MapExt <T> map;
  read_Healpix_map_from_fits(infile, map, 1, 2);
  alm.FromMap (map, lmax, lmax, niter); // get alm
  out.create  (outfile);// save them to file

  // init filter reading specfile
  arr<T> filter (lmax);
  readfitscol(filterfile, filter, 1); 
  alm.ScaleL (filter);
  
  MapExt <T> need; 
  need.FromAlm (alm, nside);// compute Healpix map
  write_Healpix_map_to_fits (out, need,planckType<T>());// save it to file

  return 0;
  
}

/////////////////////////////////////////////////
/// Compute spherical harmonic coefficients from Healpix map
/// \param infile : map file name
/// \param outfile : out alm file name
/// \param lmax : maximum order of analysis
/// \param niter : iteration number
/// \param filterfile : needlet window function
template <typename T> int need2alm (string infile, string outfile, int lmax, int niter,string filterfile)
{
  fitshandle out;
  
  AlmExt <T> alm; 
  MapExt <T> map;
  read_Healpix_map_from_fits(infile, map, 1, 2);
  alm.FromMap (map, lmax, lmax, niter); // get alm
  out.create  (outfile);// save them to file

  // init filter reading specfile
  arr<T> filter (lmax);
  readfitscol(filterfile, filter, 1); 
  alm.ScaleL (filter);
  
  write_Alm_to_fits (out,alm,alm.Lmax(),alm.Mmax(),planckType<T>());

  return 0;
  
}

/////////////////////////////////////////////////
/// Compute spherical harmonic coefficients from Healpix map
/// \param infile : alm file name
/// \param outfile : out needlet file name
/// \param lmax : maximum order of analysis
/// \param nside : needlet resolution
/// \param filterfile : needlet window function
template <typename T> int alm2need (string infile, string outfile, int lmax, int nside, string filterfile)
{
  fitshandle out;
  
  AlmExt <T> alm; 
  MapExt <T> map;

  read_Alm_from_fits (infile, alm, lmax, lmax);


  out.create  (outfile);// save them to file

  // init filter reading specfile
  arr<T> filter (lmax);
  readfitscol(filterfile, filter, 1); 
  alm.ScaleL (filter);
  
  MapExt <T> need; 
  need.FromAlm (alm, nside);// compute Healpix map
  write_Healpix_map_to_fits (out, need,planckType<T>());// save it to file

  return 0;
  
}

/////////////////////////////////////////////////
/// Generate mask 
/// \param covers : coverage fractions
/// \param kfile : reference intensity map
/// \param nside : resolution of the file
/// \param file : out file name
template <typename T> int makekpmasks (arr<T> covers, string kfile, int nside, string file)
{
  
  MapExt<T> KMap;
  read_Healpix_map_from_fits(kfile, KMap, 1, 2);

  MapExt<T> fMask;
  fMask.SetNside(nside, RING);

  KPMask (covers, fMask, KMap);

  fitshandle out;
  out.create (file);
  write_Healpix_map_to_fits (out, fMask,planckType<T>());  
}



/////////////////////////////////////////////////
/// Generate mask 
/// \param nside_in : resolution of the mask
/// \param nside_out : resolution of the file
/// \param file : out file name
template <typename T> int makegridmasks (int nside_in, int nside_out, string file)
{
  MapExt<T> Out;
  Out.SetNside(nside_out, RING);

  if (nside_in==0)
    Out.fill(0);
  else
    {
      Healpix_Map<T> Mask ;
      Mask.SetNside(nside_in, RING);
      for (int pix=0; pix<Mask.Npix(); pix++)
	Mask[pix] = pix;
      Out.Import_upgrade(Mask);
    }
  fitshandle out;
  out.create (file);
  write_Healpix_map_to_fits (out, Out,planckType<T>());
}
