/*
 * Project: Spherelib
 * Date: 09/04  
 * Authors: M. Betoule M. Le Jeune
 *
 */

/* Copyright (C) 2008 APC CNRS Université Paris Diderot <lejeune@apc.univ-paris7.fr>  
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see http://www.gnu.org/licenses/gpl.html
 */
#include "MapExt.h"
#include "window.h"
#include <gsl/gsl_rng.h>
#include <gsl/gsl_sort_double.h>
#include <gsl/gsl_linalg.h>

using namespace std;

/////////////////////////////////////////////////
/// Build a mask made of KP zones
/// \param covers : coverage fractions
/// \param Map : output zone map
/// \param KMap : reference intensity map
/////////////////////////////////////////////////
template <typename T> Status KPMask (arr<T> covers, MapExt<T> &Map, const MapExt<T> KMap)
{
  int ncovers = covers.size();
  
  if ((ncovers == 1 ) && (covers[0]==1))
    {
      Map.fill(1);
      return 0;
    }
  
  // make an histogram of reference map
  T thresh;
  arr<double> KArr (KMap.Map());
  gsl_sort (KArr.begin(), 1, KMap.Npix());
  
  int lmax = 2*KMap.Nside();
  AlmExt<T>  uKAlm (lmax,lmax); 
  arr<T> Bl (lmax+1);
  GaussBeam (Bl, 120);
  PowSpecExt<T> Beam (1, lmax);
  Beam.Set (Bl);
  
  MapExt<T>   Mask ; Mask.SetNside(KMap.Nside(), RING); Mask.fill(-1);
  MapExt<T>  dKMap ; dKMap.SetNside(64, RING);
  MapExt<T>  uKMap ; uKMap.SetNside(KMap.Nside(), RING);
  
  for (int c=0; c<ncovers; c++)
    {
      thresh = (T) KArr[(int)(ceil(covers[c]*KMap.Npix()))];
      dKMap.Import (KMap);
      for (int p=0; p<dKMap.Npix(); p++)
	if (dKMap[p]<thresh)
	  dKMap[p]=1;
	else
	  dKMap[p]=0;
      uKMap.Import(dKMap);
      uKAlm.FromMap(uKMap);
      uKAlm.ScaleL (Beam.tt());
      uKMap.FromAlm(uKAlm);
      for (int p=0; p<uKMap.Npix(); p++) 
	if ( (uKMap[p]<0.5) && (Mask[p]==-1)) 
	  Mask[p] = c; 
    }
  for (int p=0; p<uKMap.Npix(); p++) 
    if ((Mask[p]==-1)) 
      Mask[p] = ncovers; 
  Map.Import(Mask);
  for (int p=0; p<Map.Npix(); p++) 
    Map[p] = round(Map[p])+1;
  
  return 0;
}

/////////////////////////////////////////////////
/// Compute template maps up to order
/// \param Tpl : template map array
/// \param order : regression order (1 or 2)
/////////////////////////////////////////////////
template <typename T> Status getTemplate (MapExt<T> *Map, int order, int nside, Healpix_Ordering_Scheme scheme)
{
  (Map[0]).SetNside (nside, scheme);
  (Map[0]).fill(1.0);
  if (order != 1)
    {
      (Map[1]).SetNside (nside, scheme);
      (Map[2]).SetNside (nside, scheme);
      (Map[3]).SetNside (nside, scheme);
      for (int pix=0; pix<12*nside*nside; pix++)
	{
	  pointing p  =	(Map[1]).pix2ang (pix);
	  (Map[1])[pix] = sin(p.theta)*cos(p.phi);
	  (Map[2])[pix] = sin(p.theta)*sin(p.phi);
	  (Map[3])[pix] = cos(p.theta);
	}
    }
}

/////////////////////////////////////////////////
/// Remove multipole of masked map up to order
/// \param Mask : mask that will be applied
/// \param order : regression order (1 or 2)
/////////////////////////////////////////////////
template <typename T> Status Regress (MapExt<T> &Map, const MapExt<T> &Mask, int order=1, bool verbose=0)
{

  // get regression order
  int ntpl;
  if (order==1)
    ntpl = 1;
  else {
    order= 2;
    ntpl = 4;
  }
  
  // get template
  MapExt<T> Tpl [ntpl];
  getTemplate (Tpl, order, Map.Nside() , Map.Scheme());
  

  // mask * template
  MapExt<T> MskTpl [ntpl];
  for (int i=0; i<ntpl; i++){
    MskTpl[i]   = Tpl[i];
    MskTpl[i]  *= Mask;
   
  }
  
  gsl_matrix * PtPM, *iPtPM;
  gsl_permutation *p;
  int signum;
  p  = gsl_permutation_alloc (ntpl);
  PtPM  = gsl_matrix_alloc (ntpl, ntpl);
  iPtPM = gsl_matrix_alloc (ntpl, ntpl);
  for (int i=0; i<ntpl; i++){
    for (int j=i; j<ntpl; j++)
      {
	MapExt<T> tmpMap = Tpl[i];
	tmpMap *= MskTpl[j]; // template2 * mask
	T val   = tmpMap.sum();
	gsl_matrix_set (PtPM, i, j, val);
	gsl_matrix_set (PtPM, j, i, val);
      }
  }
  
  if (verbose){
    cout << "PtPM is : \n"; 
    for (int i=0; i<ntpl; i++){
      for (int j=i; j<ntpl; j++)
	{
	  cout << " " << gsl_matrix_get (PtPM, i, j);
	  cout << " " << gsl_matrix_get (PtPM, j, i);
	}}
  }

  gsl_linalg_LU_decomp (PtPM, p, &signum);
  gsl_linalg_LU_invert (PtPM, p, iPtPM);

  if (verbose){
    cout << "iPtPM is : \n"; 
    for (int i=0; i<ntpl; i++){
      for (int j=i; j<ntpl; j++)
	{
	  cout << " " << gsl_matrix_get (iPtPM, i, j);
	  cout << " " << gsl_matrix_get (iPtPM, j, i);
	}}
  }
  
  gsl_vector *PdM, *Ma;
  CBLAS_UPLO_t Uplo=CblasUpper;
  Ma  = gsl_vector_alloc (ntpl);
  PdM = gsl_vector_alloc (ntpl);

  for (int i=0; i<ntpl; i++){
    MskTpl[i] *= Map; // mask*map*template
    gsl_vector_set(PdM, i, MskTpl[i].sum());
    gsl_vector_set(Ma, i,0);
  }
  if (verbose){
    cout << "PdM is : \n"; 
    for (int i=0; i<ntpl; i++){
      cout << " " << gsl_vector_get (PdM, i);
      gsl_vector_set (Ma, i, 0);
    }}
  
  
  gsl_blas_dsymv (Uplo, 1.0, iPtPM, PdM, 1.0, Ma);
  
  if (verbose){
    cout << "Ma is : \n"; 
    for (int i=0; i<ntpl; i++){
      cout << " " << gsl_vector_get (Ma, i);
    }}
  
  for (int i=0; i<ntpl; i++){
    Map.LinComb (Tpl[i], 1.0, -1.0*gsl_vector_get (Ma, i));
  }
  
  gsl_permutation_free (p);
  gsl_vector_free (PdM);
  gsl_vector_free (Ma);
  gsl_matrix_free (PtPM);
  gsl_matrix_free (iPtPM);
  
}
