/*
 * Project: Spherelib
 * Date: 09/04  
 * Authors: M. Le Jeune, M. Betoule, F. Guilloux, G. Fay, J.-F. Cardoso
 *
 */

/* Copyright (C) 2008 APC CNRS Université Paris Diderot <lejeune@apc.univ-paris7.fr>  
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see http://www.gnu.org/licenses/gpl.html
 */
#include "lsconstants.h"
#include <gsl/gsl_blas.h>
#include <gsl/gsl_blas_types.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_sort_double.h>
#include <gsl/gsl_eigen.h>
#include <gsl/gsl_sort_vector_double.h>
#include <gsl/gsl_sort.h>
#include <gsl/gsl_sf_legendre.h>
#include <gsl/gsl_matrix_double.h>
#include "Bin.h"
using namespace std;

/////////////////////////////////////////////////
/// Return value of spline window function
/// \param x : absisses coordinate
/////////////////////////////////////////////////
double spline_value (double x)
{
  int order = 3;
  x   = 3 - 4*x;
  int    sgn = 1;
  double sumbc = 0;
  arr<double> bincoeff (order+1);
  arr<int> BC(order+1);      BC.fill(1);      BC[1] = 3;      BC[2] = 3;
  for (int c=order;c>=0;c--)
    {
      bincoeff[order-c] = ((double)BC[order-c]*sgn)/((double)(2*c+1));
      sumbc        += bincoeff[c];
      sgn  = -sgn;
    }
  double f  = bincoeff[order]/sumbc*x;
  double x2 = x*x; 
  for (int c=order-1;c>=0;c--)
    {
      x = x*x2;
      f = f + (bincoeff[c]/sumbc*x);
    }
  f = (1+f)/2;
  if (x<=-1.0)
    f = 0;
  if (x>=1.0) 
    f = 1;
  return f;
}

/////////////////////////////////////////////////
/// Compute coupling matrix
/// \param lmin : minimum multipole value
/// \param lmax : maximum multipole value
/// \param costheta : cos of the opening of spherical caps.
/////////////////////////////////////////////////
gsl_matrix * incompleteLegendreScalProd(int lmin, int lmax, double costheta){
  double pol[lmax+1+2];
  double dpol[lmax+1+2];
  double c2;
  int l1,l2, nElem, i, j;
  double  d1, d2, s1, s2;
  gsl_matrix * D;

  nElem = lmax - lmin + 1;
  
  D = gsl_matrix_alloc(nElem, nElem);
  if (D == NULL) printf("allocation failed\n");

  gsl_sf_legendre_Pl_deriv_array(lmax+1, costheta, pol, dpol);
  c2 = 1. - costheta*costheta;

  //off-diagonal terms
  for (i = 1; i < nElem; i++){
    l1 = lmin + i;
    for (j = 0; j < i; j++){
      //http://mathworld.wolfram.com/LegendrePolynomial.html (50)
      l2 = lmin + j;
      gsl_matrix_set(D, i, j, c2 * (pol[l1] * dpol[l2] - pol[l2] * dpol[l1])/ (l2*(l2+1.) - l1*(l1+1.))
		     * sqrt((2*l1+1)*(2*l2+1))*0.5); //normalisation
    } 
  }

  //diagonal terms
  //Wandelt et al 2001 equation (C20)
  //auxilliary computation : 2nd sub-diagonal and diagonal at lmin-1
  d1 = 1. - costheta;
  d1 *= 0.5;//normalisation
  d2 = (1. - costheta * costheta*costheta) / 3.;
  d2 *= 1.5;
  s2 = c2 / 2. * costheta;
  s2 *= sqrt(5)*0.5;
  for (l1 = 2; l1 < lmin+2; l1++){
    s1 = s2;
    s2 = c2 * (pol[l1+1] * dpol[l1-1] - pol[l1-1] * dpol[l1+1])/ ((l1-1.)*(l1) - (l1+1.)*(l1+2.))
      * sqrt((2*l1+3)*(2*l1-1))*0.5;//normalisation
    d1 = d2;
    //d2 = (2.*l1 - 1.)/(2.*l1+1.) * d1
    //  + (2.*l1-1.)/(2.*l1+1.) * (l1+1.)/l1 * s2
    //  - (l1-1.)/l1 * s1;
    d2 = d1
      + (2.*l1-1.)/sqrt((2.*l1-1.)*(2.*l1+3.)) * (l1+1.)/l1 * s2
      - (2.*l1+1.)/sqrt((2.*l1+1.)*(2.*l1-3.)) * (l1-1.)/l1 * s1;
  }
  

  gsl_matrix_set(D, 0, 0,d1);
  gsl_matrix_set(D, 1, 1,d2);
  for (i = 2; i < nElem - 1; i++){
    l1 = lmin + i;
    s1 = s2;
    s2 = gsl_matrix_get(D,i+1,i-1);
    
    //gsl_matrix_set(D, i, i,(2.*l1 - 1.)/(2.*l1+1.)*gsl_matrix_get(D,i-1,i-1)
    //+ (2.*l1-1.)/(2.*l1+1.) * (l1+1.)/l1 * s2
    //- (l1-1.)/l1 * s1);
    gsl_matrix_set(D, i, i,gsl_matrix_get(D,i-1,i-1)
		   + (2.*l1-1.)/sqrt((2.*l1-1.)*(2.*l1+3.)) * (l1+1.)/l1 * s2
		   - (2.*l1+1.)/sqrt((2.*l1+1.)*(2.*l1-3.)) * (l1-1.)/l1 * s1); 

  }

  i = nElem - 1;
  l1 = lmax;
  s1 = s2;
  s2 = c2 * (pol[l1+1] * dpol[l1-1] - pol[l1-1] * dpol[l1+1])/ ((l1-1.)*(l1) - (l1+1.)*(l1+2.))
    * sqrt((2*l1+3)*(2*l1-1))*0.5;//normalisation
  //gsl_matrix_set(D, i, i,(2.*l1 - 1.)/(2.*l1+1.)*gsl_matrix_get(D,i-1,i-1)
  //+ (2.*l1-1.)/(2.*l1+1.) * (l1+1.)/l1 * s2
  //- (l1-1.)/l1 * s1);
  gsl_matrix_set(D, i, i,gsl_matrix_get(D,i-1,i-1)
		 + (2.*l1-1.)/sqrt((2.*l1-1.)*(2.*l1+3.)) * (l1+1.)/l1 * s2
		 - (2.*l1+1.)/sqrt((2.*l1+1.)*(2.*l1-3.)) * (l1-1.)/l1 * s1);
  return D;
}


/////////////////////////////////////////////////
/// Set list bin, build spline functions from bands
/// \param Bands : 1 dimensional array of multipoles
/// \param SpecBin : Output bin array
/////////////////////////////////////////////////
template <typename T> Status Spline(arr<int> &Bands, arr2<T> &SpecBin){

  if (&Bands==NULL) return(7);
  //if (&m_SpecBin) { delete &m_SpecBin; }
  // Init attribut
  int m_NBand   = Bands.size()-2;
  int m_Lmin    = 0;
  int m_Lmax    = Bands[Bands.size()-1];
  
  SpecBin.alloc(m_NBand, m_Lmax+1);    
  SpecBin.fill(0.0);
  
  int lmin, lmax, maxi;
  double a1, x, f;
  for (int j=0; j<m_NBand ; j++)
    {
      lmin = Bands[j];
      lmax = Bands[j+2];
      maxi = Bands[j+1];
      
      for (int ell=lmin; ell<=lmax; ell++)
	{
	  if (ell<=maxi)
	    {
	      a1 = 0.5/(maxi-lmin);
	      x = (a1*ell)+(1-a1*maxi);
	      f = spline_value (x);
	      SpecBin[j][ell] = sqrt(1-f);
	    }
	  else
	    {
	      a1 = 0.5/(lmax-maxi);
	      x = (a1*ell)+(1-a1*lmax);
	      f = spline_value (x);
	      SpecBin[j][ell] = sqrt(f);
	    }
	}
      if (j==0)
	for (int ell=lmin; ell<=maxi; ell++)
	  SpecBin[j][ell] = 1;
    }
  
  
  return 0;
}

/////////////////////////////////////////////////
/// Return eigen vector associated to greatest eigen value
/// \param m : matrix
/// \param concentration : greatest eigen value 
/////////////////////////////////////////////////
gsl_vector * solveEigenSystem(gsl_matrix * m, double * concentration){
  gsl_vector *eval = gsl_vector_alloc (m->size1);
  gsl_matrix *evec = gsl_matrix_alloc (m->size1, m->size1);
  gsl_matrix *tmp = gsl_matrix_alloc (m->size1, m->size1);
  gsl_matrix_set_zero(tmp);
  gsl_matrix_add(tmp, m);

  gsl_eigen_symmv_workspace * w = gsl_eigen_symmv_alloc (m->size1);
  
  gsl_eigen_symmv (tmp, eval, evec, w);

  gsl_eigen_symmv_free (w);
  
  gsl_eigen_symmv_sort (eval, evec, GSL_EIGEN_SORT_VAL_DESC);

  if (concentration != NULL) *concentration = gsl_vector_get (eval, 0);
  
  gsl_vector * vec = gsl_vector_alloc (m->size1);
  gsl_vector_free (eval);
  gsl_matrix_get_col (vec, evec,0);
  gsl_matrix_free (evec);
  gsl_matrix_free (tmp);
  return vec;
  
}

/////////////////////////////////////////////////
/// Smooth coupling matrix if not positive definite
/// \param D : coupling matrix
/// \param v : eigen value vector
/// \param ev : eigen vector matrix
/// \smooth : smoothing parameter. 
/////////////////////////////////////////////////
gsl_vector * smoothing_pswf (gsl_matrix *D, double smooth=2.2204e-16)
{
  double eps = 2.2204e-16;
  double *mt = (double*) malloc (1*sizeof(double));
  double min , max;
  double fac = 1;

  // get eigen value and vectors
  int nr = D->size1;
  gsl_vector *dv; 
  dv = solveEigenSystem (D, NULL);
  gsl_sort_vector_smallest(mt, 1, dv);   min=mt[0];
  gsl_sort_vector_largest (mt, 1, dv);   max=mt[0];
  bool negative = ((min<-1.0*eps) & (max>eps));
  if (negative){
    gsl_matrix *A, *A2, *D2;
    A  = gsl_matrix_alloc (nr-2, nr);
    A2 = gsl_matrix_alloc (nr, nr);
    D2 = gsl_matrix_alloc (nr, nr);
    gsl_matrix_set_zero (A);
    gsl_matrix_set_zero (A2);
    for (int i=0; i<nr-2; i++){
      gsl_matrix_set(A, i, i, 1.0);
      gsl_matrix_set(A, i, i+1, -2.0);
      gsl_matrix_set(A, i, i+2, 1);
    }
    CBLAS_TRANSPOSE_t TransA = CblasTrans;
    CBLAS_TRANSPOSE_t TransB = CblasNoTrans;
    gsl_blas_dgemm (TransA, TransB, 1.0, A,A, 1.0, A2);
    gsl_matrix_scale(A2, smooth);
    while(negative){
      gsl_matrix_set_zero (D2);
      gsl_matrix_add  (D2 , D);
      gsl_matrix_sub  (D2 , A2);
      dv = solveEigenSystem (D2, NULL);
      gsl_sort_vector_smallest(mt, 1, dv);   min=mt[0];
      gsl_sort_vector_largest (mt, 1, dv);   max=mt[0];
      negative = ((min<-1.0*eps) && (max>eps));
      gsl_matrix_scale(A2, 10);
    }
    gsl_matrix_free (A);
    gsl_matrix_free (A2);
  }
  free(mt);
  return dv;
}

/////////////////////////////////////////////////
/// Set list bin, build prolat functions from bands
/// \param Bands : 1 dimensional array of multipoles
/// \param SpecBin : Output bin array
/////////////////////////////////////////////////
template <typename T> Status Prolate(arr<int> Bands, arr2<T> &SpecBin, arr<float> thetas, int band2deg=0){

  if (&Bands==NULL) return(7);
 
  
  //if (&m_SpecBin) { delete &m_SpecBin; }
  // Init attribut
  int m_NBand   = Bands.size()-2;
  int m_Lmin    = 0;
  int m_Lmax    = Bands[Bands.size()-1];

  SpecBin.alloc(m_NBand, m_Lmax+1);    
  SpecBin.fill(0.0);
  
  int lmin, lmax;
  T theta, solid, lmean, delta, z;
  double *mt = (double*) malloc (1*sizeof(double));
  for (int j=m_NBand-1; j>=0 ; j--)
    {
      lmin  = Bands[j];
      lmax  = Bands[j+2];
      lmean = (lmin+lmax)/2.0;
      delta = (lmax-lmin)/2.0;
      solid = 4*pi/ ( ((lmax+1)*(lmax+1)) - (lmin*lmin) );
      if (band2deg>0)
	switch (band2deg){
	  //jfc
	case 1 : 
	  theta = acos ( 1 - solid /(2*pi));  
	  theta = theta*180/pi;
	  break;
	  //gf
	case 2:
	  theta = 1 / sqrt(lmean);
	  theta = theta*180/pi;
	  break;
	  //jd
	case 3: 
	  theta = 1/lmean>1/(2*delta) ? 1/lmean : 1/(2*delta);
	  theta = theta*180/pi;
	}
      else
	theta = (T)thetas[j];
      z     = cos(theta*pi/180);
      gsl_matrix *D;
      D = incompleteLegendreScalProd(lmin, lmax, z);
      int nr = D->size1;
      for (int i=0; i<nr; i++)
	for (int j=i; j<nr; j++)
	  gsl_matrix_set(D, i, j, gsl_matrix_get(D, j, i));
      
      gsl_vector *v, *dv;  // eigen value
      gsl_matrix *ev;
      dv = gsl_vector_alloc (nr);
      ev = gsl_matrix_alloc (nr, nr);
      v  = smoothing_pswf (D );// set and sort eigen vectors
      
      // ev contains normalization matrix
      gsl_matrix_set_zero (ev);
      gsl_vector_set_zero (dv);
      for (int ell=0; ell<nr; ell++)
	gsl_matrix_set (ev, ell, ell, sqrt(8*pi*pi/(2*(lmin+ell)+1)));
      CBLAS_UPLO_t Uplo=CblasUpper;
      gsl_blas_dsymv (Uplo, 1, ev, v, 1, dv);
      
      double min , max;
      gsl_sort_vector_smallest(mt, 1, dv);   min=(mt[0]);
      gsl_sort_vector_largest (mt, 1, dv);   max=(mt[0]);
      if (max <0)
	max = abs(min);
      bool changesign = (abs(min)>=max);
      for (int ell=lmin; ell<=lmax; ell++)
	{
	  double f= gsl_vector_get(dv, ell-lmin)/max;
	  f = changesign ? -1.0*f : f;
	  SpecBin[j][ell] = f;
	}
      gsl_matrix_free (ev);
      gsl_vector_free (v);
      gsl_vector_free (dv);
    }
  free(mt);
  return 0;
}

template <typename T> Status GaussBeam (arr<T> &Bl, double fwhm_arcmin)
{
  int lmax = Bl.size()-1;
  double fwhm_rad = (fwhm_arcmin*2*pi/60)/360; //convert to radian
  for (int ell=0; ell<=lmax; ell++)
    Bl[ell] = sqrt(exp((double)-ell*(ell+1)/8/log(2) * (fwhm_rad*fwhm_rad))) ;
  return 0;
}
