/*
 * Project: Spherelib
 * Date: 08/10  
 * Authors: J.M. Colley - M. Le Jeune
 *
 */

/* Copyright (C) 2008 APC CNRS Université Paris Diderot <lejeune@apc.univ-paris7.fr>  
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see http://www.gnu.org/licenses/gpl.html
 */

#define ALM_DBG

#include <stdlib.h>
#include <string.h>
#include "CD_CommonDefinition.h"
#include "xcomplex.h"
#include "alm.h" 
#include "healpix_base.h"
#include "fitshandle.h"
#include "healpix_data_io.h"
#include "AlmExt.h"


using namespace std; 

/******************************************************************************
 * Print alm values 
 *
 */
template <typename T> 
void AlmExt<T>::PrintAlm(long lmax,long mmax,const string &title) 
{
  cout << "\n"<<title; // print title of the alm
  // check on Alm consistency
  if ((this->Lmax() ==0) && (this->Mmax() ==0)) {
    CD_WarningStatus(1);
    return;
  }
  if ( (lmax < 0) || (mmax < 0))
    { CD_WarningStatus(2);
      return;
    }
  if ( (lmax > this->Lmax() ) ||(mmax > this->Mmax())       )        
    {   
      char Mes[500];
      sprintf(Mes,"\nLmax = %d - %ld , Mmax = %d - %ld \n", this->Lmax(),lmax, this->Mmax(),mmax); 
      CD_WarningStatusMes(3, Mes);
      if (lmax > this->Lmax())
	lmax = this->Lmax();
      if (mmax > this->Mmax())
	mmax = this->Mmax();
    }

  //print alm value for each l and m up to lmax and mmax
  for(long Myl=0; Myl <= lmax; Myl++) 
    {
      for(long Mym=0; Mym <= Myl; Mym++)
	cout <<"["<<Myl<<","<<Mym<<"]="<<(*this)(Myl,Mym)<<" "; 
      cout<<"\n";
    }    
  return;
}
template void AlmExt<double>::PrintAlm(long lmax,long mmax,const string &title) ;
template void AlmExt<float>::PrintAlm(long lmax,long mmax,const string &title) ;

/******************************************************************************
 * Print attribut of class Amm
 *
 */
template <typename T> 
void AlmExt<T>::PrintAttribut(const string &Name) 
{
    cout << "\nAlmExt object Name :'" << Name << "'";
    cout << "\n  Lmax    : " << this->Lmax();
    cout << "  Mmax    : " << this->Mmax();
    cout << "\n  PathWeight: '" << m_PathWeight << "'";
    cout << "\n  Status  : " << m_Status; 
    //cout << "\n  & arr   : " << this->mstart(); 
    PrintAlm(3, 3, "  First Alm values: \n");
}
template void AlmExt<double>::PrintAttribut(const string &Name) ;
template void AlmExt<float>::PrintAttribut(const string &Name) ;



/******************************************************************************
 *  
 * ScaleL exist also in base class with arr<T> 
 * //CD_return( (*this).ScaleL( (double *)&(Beam.tt()[0]) ) );  
 * Sept 2007, Jmc
 */



/******************************************************************************
 *
 * Compute the linear combination AlmIntern <- alpha.AlmIntern + beta.AlmB 
 *
 */
template <typename T> 
Status AlmExt<T>::LinComb(const AlmExt<T> &AlmB, T alpha,  T beta )
{
  long Lmax = this->Lmax();
  if (Lmax > AlmB.Lmax())
    {
      char Mes[500];
      sprintf(Mes,"\nLmax = %ld - %ld\n", Lmax,AlmB.Lmax()); 
      CD_WarningStatusMes(3, Mes);
      Lmax = AlmB.Lmax();	
    } 
  
  this->Scale(alpha);
  
  for (long ell=0; ell<=Lmax; ell++)
    for (long emm=0; emm <=ell; emm++)
      (*this)(ell, emm) += beta*AlmB(ell,emm);
  
  CD_returnOK;
}
template Status AlmExt<double>::LinComb(const AlmExt<double> &AlmB, double alpha,  double beta );
template Status AlmExt<float>::LinComb(const AlmExt<float> &AlmB, float alpha,  float beta );

/******************************************************************************
 *
 * Compute the linear combination AlmIntern <- AlmIntern + beta.AlmB
 *
 */
template <typename T> 
Status AlmExt<T>::LinComb(const AlmExt<T> &AlmB,T beta )
{
  (*this).LinComb(AlmB, 1, beta);
  CD_returnOK;
}
template Status AlmExt<double>::LinComb(const AlmExt<double> &AlmB,double beta );
template Status AlmExt<float>::LinComb(const AlmExt<float> &AlmB,float beta );

/******************************************************************************
 *
 */
template <typename T>
AlmExt<T>& AlmExt<T>::operator=(const AlmExt<T>& OtherAlm)
{
    if (this != &OtherAlm)
    {
      this->Set(OtherAlm.Lmax(), OtherAlm.Mmax());
      for (long Lidx=0; Lidx <= this->Lmax() ; Lidx++)
	for (long Midx=0 ; Midx <= Lidx ; Midx++)
	  (*this)(Lidx, Midx) =  OtherAlm(Lidx, Midx);
      
      this->m_PathWeight = OtherAlm.m_PathWeight;
    }
    return *this;
}

template AlmExt<double>& AlmExt<double>::operator=(const AlmExt<double>& OtherAlm);
template AlmExt<float>& AlmExt<float>::operator=(const AlmExt<float>& OtherAlm);


