/*
 * Project: Spherelib
 * Date: 08/10  
 * Authors: J.M. Colley - M. Le Jeune
 *
 */

/* Copyright (C) 2008 APC CNRS Université Paris Diderot <lejeune@apc.univ-paris7.fr>  
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see http://www.gnu.org/licenses/gpl.html
 */

#ifndef BIN_H
#define BIN_H

#include <string.h>
#include "CD_CommonDefinition.h"
#include <assert.h>
#include <stdio.h>

using namespace std;

/////////////////////////////////////////////////
/// Bin contains the description of the spectral binning
/////////////////////////////////////////////////
class Bin
{
 public:
  
  /////////////////////////////////////////////////
  /// Default constructor    
  Bin() {InitAttribut();}       
  
  /////////////////////////////////////////////////
  /// Constructor allocation memory  
  Bin (arr2<double> &Bins)  {
    CD_ExitIfNok( Set(Bins) );
  }
  
  /////////////////////////////////////////////////
  /// Destructor : 
  virtual ~Bin() { 
    //if (&m_SpecBin) { delete &m_SpecBin; }
  }
    
    
  
  // ====
  // FUNC
  // ====
  
  
  /////////////////////////////////////////////////
  /// Return status of last call method    
  Status GetStatus() {return(m_Status);}
  
  ////////////////////////////////////////////////
  /// Return number of spectral bins
  long GetNBand () {return(m_NBand);}
  
  ////////////////////////////////////////////////
  /// Return global maximum value of L
  long GetLmax () {return(m_Lmax);}
  
  ////////////////////////////////////////////////
  /// Return global minimum value of L
  long GetLmin () {return(m_Lmin);}

  /////////////////////////////////////////////////
  /// Print attribut of class
  /// \param Name : title of the alm
  /////////////////////////////////////////////////
  void PrintAttribut(string Name){
      cout << "\nBinObject:'" << Name << "'";
      cout << "\n  Lmin   : " << m_Lmin;
      cout << "\n  Lmax   : " << m_Lmax;
      cout << "\n  Status : " << m_Status; 
      cout << "\n  NBand  : " << m_NBand;
  }
  
  /////////////////////////////////////////////////
  /// Set list bin with Array2D
  /// \param Bins : 2 dimensions array
  /////////////////////////////////////////////////
  Status Set(arr2<double> &Bins){
      if (&Bins==NULL) CD_return(7);

      //if (&m_SpecBin) { delete &m_SpecBin; }

      // Init attribut
      m_NBand   = Bins.size1();
      m_Lmin    = 0;
      m_Lmax    = Bins.size2()-1;
      m_SpecBin = Bins;    
      CD_returnOK;
  }


  /////////////////////////////////////////////////
  /// Return the greatest non zero ell value of qth bin
  /// \param q : bin number
  /////////////////////////////////////////////////
  long getLmax (long q)
  {
    assert(&m_SpecBin);
    long Lmax =  m_Lmax;
    for (long ell=m_Lmax; ell>=0 ; ell--)
      if (m_SpecBin[q][ell] > 0.0)
	{
	  Lmax = ell;
	  break;
	}
    return Lmax;
  }
  
  
  /////////////////////////////////////////////////
  /// Return the smallest non zero ell value of qth bin
  /// \param q : bin number
  /////////////////////////////////////////////////
  long getLmin (long q)
  {
    assert(&m_SpecBin);
    long Lmin= 0;
    for (long ell=0; ell<=m_Lmax ; ell++)
      if (m_SpecBin[q][ell] != 0)
	{
	  Lmin = ell;
	  break;
	}
    return Lmin;
  }
  
  /////////////////////////////////////////////////
  /// Return the number of bins
  /////////////////////////////////////////////////
  long getNBand () {return m_SpecBin.size1();}
  
  long getLmax () {return m_SpecBin.size2()-1;}
  
  
  /////////////////////////////////////////////////
  /// Return the best Healpix resolution for qth bin (depends on lmax)
  /// \param q : bin number
  /////////////////////////////////////////////////
  long getNside (long q)
  {
    long Lmax  = getLmax(q);
    long Nside;
    long n = (long) ceil(log2( Lmax/2.0 ));
    Nside = (long) pow((double) 2,(double) n) ;
    return Nside;
  }
  
  /////////////////////////////////////////////////
  /// Return the window function value of the qth bin (vector of lmax+1 values)
  /// \param q : bin number
  /////////////////////////////////////////////////
  arr<double>* getQFilter (long q)
  { 
    long  Lmax    = m_SpecBin.size2();
    arr<double>* F = new arr<double>(Lmax+1);
    for (long ell=0; ell<=Lmax; ell++)
      (*F)[ell] = m_SpecBin[q][ell];
    return F;
  }

  /////////////////////////////////////////////////
  /// Return window functions 
  /////////////////////////////////////////////////
  const arr2<double> &getBins () const    {      return m_SpecBin;    }
  
  /////////////////////////////////////////////////
  /// Print min and max values of bins
  /// \param QBegin : first bin print 
  /// \param QEnd   : last bin print 
  /// \param title  : title of the Bin
  /////////////////////////////////////////////////
  Status PrintBin(long QBegin, long QEnd, string title="")
  {
    
    Bin::PrintAttribut(title);
    cout << "\n  SpecBin: " << &m_SpecBin;
    
    if (&m_SpecBin == NULL) CD_returnOK; 
    
    long Lmin, Lmax; 
    
    for (long q=QBegin; q<=  QEnd; q++){
      Lmin = getLmin(q);
      Lmax = getLmax(q);
      printf("\n(Line=%ld) = l=%ld...%ld ",q, Lmin,Lmax); 
      for (long Li=Lmin ;  Li <=Lmax; Li++) printf("%g ",m_SpecBin[q][Li] );
    }
    
    cout << "\n";
    CD_returnOK;
  }
  
  /////////////////////////////////////////////////
  /// Print min and max values of bins
  /// \param title  : title of the Bin
  /////////////////////////////////////////////////
  Status PrintBin(string title="") {
    CD_return( PrintBin(0, m_NBand-1, title));
  }
  
  ////////////////////////////////////////////////
  /// Return number of modes of each bin (sum window weighted by 2l+1)
  virtual double* GetNbMode ();
  
  ////////////////////////////////////////////////
  /// Unbin power spectrum
  /// \param Spec   : binned spectrum
  virtual arr<double>* UnbinSpec (arr<double>* Spec);
  
  
  ////////////////////////////////////////////////
  /// Rescale window function array to Lmax by removing unnecessary scales 
  /// \param Lmax   : highest multipole value
  Status Rescale (long Lmax);
  
  
  
 protected :
  
  // ====
  // FUNC
  // ====
  /////////////////////////////////////////////////
  /// initialisation data member (or attribut) of object
  void InitAttribut() {
    m_Status=0; 
    m_Lmin=0;
    m_Lmax=0;
    m_NBand = 0;
  }
  // ====
  // DATA
  // ====
  
  /// Description of the spectral bins
  /// m_SpecBin[0..(NbBand-1), 0..Lmax]
  arr2<double> m_SpecBin;
  
  /// Number of spectral bins
  long m_NBand;

  /// l min
  long m_Lmin;

  /// l max
  long m_Lmax; 
  
  /// Identify error of last call method, O is Ok
  Status m_Status;

};

#endif
