/*
 * Project: Spherelib
 * Date: 08/11  
 * Authors: J.M. Colley - M. Le Jeune
 *
 */

/* Copyright (C) 2008 APC CNRS Université Paris Diderot <lejeune@apc.univ-paris7.fr>  
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see http://www.gnu.org/licenses/gpl.html
 */

#ifndef COVMAT_H
#define COVMAT_H

#include "CD_CommonDefinition.h"
#include <string>
#include "healpix_map.h"
#include <gsl/gsl_blas.h>
#include <gsl/gsl_permutation.h>
#include <gsl/gsl_linalg.h>

template <typename T> class PowSpecExt;
template <typename T> class AlmExt;
template <typename T> class MapExt;

using namespace std; 


/////////////////////////////////////////////////
/// CovMat stands for a set of covariance matrices
/////////////////////////////////////////////////
template<typename T> 
class CovMat 
{
 public:
  
  /////////////////////////////////////////////////
  /// Default constructor
  /////////////////////////////////////////////////
  CovMat() {InitAttribut();}
  

  /////////////////////////////////////////////////
  /// Destructor
  /////////////////////////////////////////////////
  virtual ~CovMat() {}
  

  // ====
  // FUNC
  // ====
  
  // Setter 
  /////////////////////////////////////////////////
  /// Set covariance matrices
  /// \param covmat : covariance matrices array
  Status setCovMat (arr2<T> covmat)
    {
      m_NbBin = covmat.size2();
      m_NbCross = covmat.size1();
      m_NbIn   = (int) floor(sqrt(m_NbCross*2)); 
      m_CovMat = covmat;
      CD_returnOK;
    }

  /////////////////////////////////////////////////
  /// Set number of modes
  /// \param nbmode : modes number array
  Status setNbMode (arr<T> nbmode)
    {
      if (m_NbBin!= nbmode.size())
	CD_return(17);

      m_NbMode = nbmode;
      CD_returnOK;
    }
   

  // Getter
  /////////////////////////////////////////////////
  /// Return status of last call method    
  Status GetStatus() {return(m_Status);}

  /////////////////////////////////////////////////
  /// Return covariance matrices
  /////////////////////////////////////////////////
  const arr2<T> &getCovMat () const    {      return m_CovMat;    }
  
  /////////////////////////////////////////////////
  /// Return number of modes
  /////////////////////////////////////////////////
  const arr<T> &getNbMode () const { return m_NbMode; }

  /////////////////////////////////////////////////
  /// Return number of matrix entries
  /////////////////////////////////////////////////
  int getNbIn () {return m_NbIn;}

  ///////////////////////////////////////////////// 	 
  /// Return number of matrices 	 
  ///////////////////////////////////////////////// 	 
  int getNbBin () {return m_NbBin;}
  

  /////////////////////////////////////////////////
  /// constructor : 
  /// \param Maps : collection Healpix maps
  /// \param NbMap : number of maps
  /// \param Mask : Healpix map which sets spatial zones on which average is performed
  /////////////////////////////////////////////////
  template <typename T2>Status FromMap (MapExt<T> *Maps, long NbMap, MapExt<T2> &Mask)
  {
    // check number of maps
    if (NbMap>0)
      m_NbIn = NbMap;
    else
      CD_return(12);
    
    // get number of zones
    T2 min , max;
    Mask.minmax(min,max);
    m_NbBin  = (long) max;
    
    // set size of covmat
    m_NbCross = m_NbIn*(m_NbIn+1)/2;
    m_CovMat.alloc (m_NbCross, m_NbBin); 
    m_NbMode.alloc (m_NbBin);
    m_CovMat.fill(0);
   
    // check nside and ordering compatibility
    for (int m1=0; m1<NbMap; m1++)
      CD_iCheck( Maps[m1].Conformable(Mask) );

    int i  = 0;           
    for (int m1=0; m1<NbMap; m1++)
      for (int m2=0; m2<=m1; m2++)
	{
	  m_NbMode.fill(0);
#pragma omp parallel
	  {
	    // cross term index 
	    T tempwei[m_NbBin];
	    T tempval[m_NbBin];
	    for (long z=0 ; z<m_NbBin; z++){
	      tempwei[z] = 0;
	      tempval[z] = 0;
	    }
/* 	    arr<T> tempwei (m_NbBin); */
/* 	    arr<T> tempval (m_NbBin); // sum pixel by pixel product for each cross term */
/* 	    tempwei.fill(0); */
/* 	    tempval.fill(0); */
	    
#pragma omp for
	    for (long pix=0; pix<Mask.Npix() ; pix++)
	      {
		tempval[(int)Mask[pix]-1] += (Maps[m1])[pix] * (Maps[m2])[pix];// sum product  
		tempwei[(int)Mask[pix]-1] += 1;
	      }
#pragma omp critical
	    {
	      //i = (int)((m1+1)*(m1+2)/2)-1;
	      for (long z=0 ; z<m_NbBin; z++)
		{
		  m_NbMode[z]  += tempwei[z];
		  m_CovMat[i][z] += tempval[z];
		}
	    }
	  }
	  for (long z=0 ; z<m_NbBin; z++)
	    m_CovMat[i][z] /= m_NbMode[z];  // weight stats with number of pixel
	  
	  i++;

	}
    
    CD_returnOK;
  }

  
  /////////////////////////////////////////////////
  /// constructor : 
  /// \param Maps : collection Healpix maps
  /// \param NbMap : number of maps
  /// \param Mask : Healpix map which sets spatial zones on which average is performed
  /////////////////////////////////////////////////
  template <typename T2>Status FromMapMinusMean (MapExt<T> *Maps, long NbMap, MapExt<T2> &Mask)
    {
      // check number of maps
      if (NbMap>0)
	m_NbIn = NbMap;
      else
	CD_return(12);
      
      // get number of zones
      T2 min , max;
      Mask.minmax(min,max);
      m_NbBin  = (long) max;
      
      // set size of covmat
      m_NbCross = m_NbIn*(m_NbIn+1)/2;
      m_CovMat.alloc (m_NbCross, m_NbBin); 
      m_NbMode.alloc (m_NbBin);
      m_CovMat.fill(0);
      arr2<T> Vmean (m_NbIn, m_NbBin);
      Vmean.fill(0);

      // check nside and ordering compatibility
      for (int m1=0; m1<NbMap; m1++)
	CD_iCheck( Maps[m1].Conformable(Mask) );
      
      for (int m1=0; m1<NbMap; m1++)
	{
	  m_NbMode.fill(0);
#pragma omp parallel
	  {
	    // cross term index 
	    T tempwei[m_NbBin];
	    T tempval[m_NbBin];
	    for (long z=0 ; z<m_NbBin; z++){
	      tempwei[z] = 0;
	      tempval[z] = 0;
	    }
#pragma omp for
	    for (long pix=0; pix<Mask.Npix() ; pix++)
	      {
		tempval[(int)Mask[pix]-1] += (Maps[m1])[pix];
		tempwei[(int)Mask[pix]-1] += 1;
	      }
#pragma omp critical
	    {
	      //i = (int)((m1+1)*(m1+2)/2)-1;
	      for (long z=0 ; z<m_NbBin; z++)
		{
		  m_NbMode[z]  += tempwei[z];
		  Vmean[m1][z] += tempval[z];
		}
	    }
	  }
	  for (long z=0 ; z<m_NbBin; z++)

	    Vmean[m1][z] /= m_NbMode[z];  // weight stats with number of pixel
	}

      int i  = 0;           
      for (int m1=0; m1<NbMap; m1++)
	for (int m2=0; m2<=m1; m2++)
	  {
	    m_NbMode.fill(0);
#pragma omp parallel
	    {
	      // cross term index 
	      T tempwei[m_NbBin];
	      T tempval[m_NbBin];
	      for (long z=0 ; z<m_NbBin; z++){
		tempwei[z] = 0;
		tempval[z] = 0;
	      }
	      /* 	    arr<T> tempwei (m_NbBin); */
	      /* 	    arr<T> tempval (m_NbBin); // sum pixel by pixel product for each cross term */
	      /* 	    tempwei.fill(0); */
	      /* 	    tempval.fill(0); */
	      
#pragma omp for
	      for (long pix=0; pix<Mask.Npix() ; pix++)
		{
		  tempval[(int)Mask[pix]-1] += ((Maps[m1])[pix]-Vmean[m1][(int)Mask[pix]-1]) * ((Maps[m2])[pix]-Vmean[m2][(int)Mask[pix]-1]);// sum product  
		  tempwei[(int)Mask[pix]-1] += 1;
		}
#pragma omp critical
	      {
		//i = (int)((m1+1)*(m1+2)/2)-1;
		for (long z=0 ; z<m_NbBin; z++)
		  {
		    m_NbMode[z]  += tempwei[z];
		    m_CovMat[i][z] += tempval[z];
		  }
	      }
	    }
	    for (long z=0 ; z<m_NbBin; z++)
	      m_CovMat[i][z] /= m_NbMode[z];  // weight stats with number of pixel
	    
	    i++;
	    
	  }
      
      CD_returnOK;
    }
  
  
  
  

  ///////////////////////////////////////////////// 
  /// constructor : 	   /// constructor :
  /// \param Maps : collection Healpix maps 	 
  /// \param NbMap : number of maps 	 
  /// \param fwhm : Gaussian filter 	 
  /// \param lmax : maximum order of analysis 	 
  ///////////////////////////////////////////////// 	 
  Status FromMap (MapExt<T> *Maps, int NbMap, double fwhm, int lmax, int nside) 	 
  { 	 
    // check number of maps 	 
    if (NbMap>0) 	 
      m_NbIn = NbMap; 	 
    else 	 
      CD_return(12); 	 
  	 
    // check nside and ordering compatibility 	 
    for (int m1=1; m1<NbMap; m1++) 	 
      CD_iCheck( Maps[m1].Conformable(Maps[0]) ); 	 
    
    m_NbBin  = 12*nside*nside; 	 
    
    // set size of covmat 	 
    m_NbCross = m_NbIn*(m_NbIn+1)/2; 	 
    m_CovMat.alloc (m_NbCross, m_NbBin); 	 
    m_NbMode.alloc (m_NbBin); 	 
    m_CovMat.fill(0); 	 
    
    arr<T> Gaussbeam (lmax+1); 	 
    GaussBeam (Gaussbeam, fwhm); 	 
    int i  = 0; 	 
    for (int m1=0; m1<NbMap; m1++) 	 
      for (int m2=0; m2<=m1; m2++) 	 
	{ 	 
	  MapExt<T> tmp (Maps[m1].Map(), Maps[m1].Scheme()); 	 
	  tmp *= Maps[m2]; 	 
	  
	  AlmExt<T> alm (lmax, lmax); 	 
	  alm.FromMap(tmp); 	 
	  alm.ScaleL (Gaussbeam); 	 
	  
	  tmp.FromAlm(alm , nside, Maps[m1].Scheme()); 	 
	  for (int pix=0; pix<m_NbBin ; pix++) 	 
	    m_CovMat[i][pix] = tmp[pix]; 	 
	  i++; 	 
	} 	 
    int w = (int)floor( (fwhm*(Maps[0].Nside()/2)*pi) / (2*60*180)); 	 
    for (int pix=0; pix<m_NbBin ; pix++) 	 
      m_NbMode[pix] = w*w; 	 
    
    CD_returnOK; 	 
  }
  /////////////////////////////////////////////////
  /// constructor : 
  /// \param Alms : collection Healpix alms
  /// \param NbMap : number of maps
  /////////////////////////////////////////////////
  Status FromAlm (AlmExt<T> *Alms, long NbMap, double fsky=1.0)
  {
    // check number of maps
    if (NbMap>0)
      m_NbIn = NbMap;
    else
      CD_return(12);
    
    // get number of bins
    int lmax = Alms[0].Lmax();
    m_NbBin  = lmax+1;
    
    // set size of covmat
    m_NbCross = m_NbIn*(m_NbIn+1)/2;
    m_CovMat.alloc (m_NbCross, m_NbBin); 
    m_NbMode.alloc (m_NbBin);
    m_CovMat.fill(0);
    m_NbMode.fill(0);

    // check compatibility
    for (int m1=1; m1<NbMap; m1++)
      CD_iCheck( Alms[m1].Conformable(Alms[0]) );

    int i  = 0;                 // cross term index 
    PowSpecExt<T> p (1, lmax);

    for (int m1=0; m1<NbMap; m1++)
      for (int m2=0; m2<=m1; m2++)
	{
	  p.FromAlm (Alms[m1], Alms[m2]);
	  
	  for (long ell=0; ell<m_NbBin ; ell++)
	    {
	      m_CovMat[i][ell] = p.tt(ell)/fsky;
	      m_NbMode[ell] = (2*ell+1)*fsky; // save weight
	    }
	  i++;
	}
    CD_returnOK;
   }
   
  
  
  

  

  void PrintCovMat(long begin, long end, const string &title) ;


protected :

  /// initialisation data member or attribut of object
  void InitAttribut() {m_Status=0; m_NbIn = 0; m_NbBin = 0; m_NbCross = 0;}
  
  /// Identify error of last call method, O is Ok
  Status m_Status;

  long m_NbIn ;

  long m_NbBin ;

  long m_NbCross;

  arr2<T> m_CovMat;
  
  arr<T> m_NbMode;

};



#endif
