/*
 * Project: Spherelib
 * Date: 08/10  
 * Authors: J.M. Colley - M. Le Jeune
 *
 */

/* Copyright (C) 2008 APC CNRS Université Paris Diderot <lejeune@apc.univ-paris7.fr>  
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see http://www.gnu.org/licenses/gpl.html
 */

#ifndef MAPEXT_H
#define MAPEXT_H

#include "healpix_map.h"
#include "CD_CommonDefinition.h"
#include "alm_healpix_tools.h"
#include <omp.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_sort_double.h>
#include <stdio.h>

using namespace std;

// circular dependency between AlmExt and MapExt
// anticipated declaration of AlmExt
template <typename T> class AlmExt;
template <typename T> class PowSpecExt;

/////////////////////////////////////////////////
/// MapExt is an extension of the Healpix class healpix_map
/////////////////////////////////////////////////
template <typename T>
class MapExt : public Healpix_Map<T>
{
 public:
  
  /////////////////////////////////////////////////
  /// Default constructor
  /////////////////////////////////////////////////
  MapExt() : Healpix_Map<T>() {InitAttribut();}
  
  /////////////////////////////////////////////////
  /// constructor : load all the pixel of the map in memory
  /// \param data : is the collection of pixel values of the map
  /// \param scheme : is the Healpix ordering (could be NESTED or RING)
  /////////////////////////////////////////////////
  MapExt(const arr< T > &data, Healpix_Ordering_Scheme scheme) : 
    Healpix_Map<T> (data, scheme) {InitAttribut();}
  
  /////////////////////////////////////////////////
  /// Destructor
  /////////////////////////////////////////////////
  virtual ~MapExt() {}
  
  
  // ====
  // FUNC
  // ====
  
  // Getter
  /////////////////////////////////////////////////
  /// Return status of last call method    
  Status GetStatus() {return(m_Status);}
  
  /////////////////////////////////////////////////
  /// Replace undef pixel with 0 value
  /////////////////////////////////////////////////
  Status SetUndefTo0 ()
  {
    long CptPixel = 0;
    for(long Ls=0; Ls < this->npix_ ; Ls++) {
      if (approx<T>((*this)[Ls], Healpix_undef)){
	(*this)[Ls] = 0;
	CptPixel++;
      }        
    }
    if (CptPixel) fprintf(stdout,"\nMove to 0 Healpix default value for %ld pixels", CptPixel);
    CD_returnOK;
  }
  
  // Linear combination
  /////////////////////////////////////////////////
  /// Compute the linear combination MapIntern <- alpha.MapIntern + beta.MapB
  /// \param MapB : map to combine with
  /// \param alpha : coefficient applied to the intern map
  /// \param beta : coefficient applied to MapB
  /////////////////////////////////////////////////
  template<typename T2> Status LinComb(const MapExt<T> &MapB,T2 alpha,T2 beta)
    {
      CD_iCheck( Conformable(MapB) ); // check that Nside and Ordering of both maps match
      for(long idx=0; idx < this->npix_; idx++)    
	(*this)[idx] = alpha*((*this)[idx]) + beta*MapB[idx];
      CD_returnOK;
    }
  
  /////////////////////////////////////////////////
  /// Compute the linear combination MapIntern <- MapIntern + beta.MapB 
  /// \param MapB : map to combine with
  /// \param beta : coefficient applied to MapB
  /////////////////////////////////////////////////
  template<typename T2> Status LinComb(const MapExt<T> &MapB, T2 beta)
    {
      CD_iCheck( Conformable(MapB) );// check that Nside and Ordering of both maps match
      for(long idx=0; idx < this->npix_; idx++)    
	(*this)[idx] += beta*MapB[idx];
      CD_returnOK;
    }
  
  /////////////////////////////////////////////////
  /// Compute the linear combination MapIntern <- MapIntern + beta.MapB over pixel domains
  /// \param MapB : map to combine with
  /// \param beta : coefficients applied to MapB 
  /// \param Mask : pixel domains
  /////////////////////////////////////////////////
  template<typename T2> Status LinCombZone(const MapExt<T> &MapB, T* beta, MapExt<T2> &Mask)
    {
      CD_iCheck( Conformable(MapB) );// check that Nside and Ordering of both maps match
      CD_iCheck( Conformable(Mask) ); 
      T2 min, max;
      Mask.minmax(min,max);
      long nzone = (long) max;
      // for each pixel if zone z
      for (int pix=0; pix<Mask.Npix(); pix++)
	(*this)[pix] += beta[(int)Mask[pix]-1]*MapB[pix];
      
      
      CD_returnOK;
    }

  ///////////////////////////////////////////////// 	 
  /// Compute the linear combination MapIntern <- MapIntern + beta.MapB 	 
  /// \param MapB : map to combine with 	 
  /// \param beta : coefficients applied to MapB 	 
  ///////////////////////////////////////////////// 	 
  Status LinCombZone(const MapExt<T> &MapB, const T* beta) 	 
  { 	 
    CD_iCheck( Conformable(MapB) );// check that Nside and Ordering of both maps match 	 
    long nzone = MapB.Npix(); 	 
    // for each pixel if zone z 	 
    for (int pix=0; pix<nzone; pix++) 	 
      (*this)[pix] += beta[pix]*MapB[pix]; 	 
    
    
    CD_returnOK; 
  }
  

  /////////////////////////////////////////////////
  /// Compute the mean over pixel domains
  /// \param Mean : output mean vector
  /// \param Mask : pixel domains
  /////////////////////////////////////////////////
  template<typename T2> Status MeanZone(arr<T> &Mean, MapExt<T2> Mask)
    {
      CD_iCheck( Conformable(Mask) ); 
      T2 min, max;
      Mask.minmax(min,max);
      long nzone = (long) max;
      Mean.alloc(nzone);
      Mean.fill(0);
      arr<int>npix(nzone);
      npix.fill(0);

#pragma omp parallel
      {
#pragma omp for
	// for each pixel if zone z
	for (int pix=0; pix<Mask.Npix(); pix++){
	  Mean[Mask[pix]-1] += (*this)[pix];
	  npix[Mask[pix]-1]++;}
      }
      for (int z=0; z<nzone; z++)
	Mean[z] = Mean[z]/npix[z];
      
      CD_returnOK;
    }
  
  /////////////////////////////////////////////////
  /// Compute the linear combination MapIntern <- c1.Map1 +  + cn.Mapn  over pixel domains
  /// if empty object, load it, compute and free it
  /// \param ListMap : collection of maps to combine
  /// \param ListCoef : collection of coefficient to applied to the maps
  /// \param NbMap : number of maps
  /// \param Mask : pixel domains
  /////////////////////////////////////////////////
  template<typename T2> Status SetWithLinCombZone(const MapExt<T> **ListMap,
						  const arr2<T> ListCoef,
						  long NbMap,MapExt<T2> &Mask)
    {
      CD_iCheck( Conformable(Mask) ); 
      T2 min, max;
      Mask.minmax(min,max);
      long nzone = (long) max;
      // for each pixel if zone z
      for (int pix=0; pix<Mask.Npix(); pix++)
	for(long Idx=0;Idx < NbMap; Idx++) 
	  (*this)[pix] += ListCoef[Idx][Mask[pix]-1]* (*ListMap[Idx])[pix];
      
      CD_returnOK;
    }

  /////////////////////////////////////////////////
  /// Compute the linear combination MapIntern <- c1.Map1 +  + cn.Mapn 
  /// if empty object, load it, compute and free it
  /// \param ListMap : collection of maps to combine
  /// \param ListCoef : collection of coefficient to applied to the maps
  /// \param NbMap : number of maps
  /////////////////////////////////////////////////
  template<typename T2> Status SetWithLinComb(const MapExt<T> **ListMap,
					      const T2 *ListCoef,
					      long NbMap)
    {
      MapExt<T> TempMap; // temp map object used to load the different maps
      //check on number of maps
      if (NbMap > 0) {
	*this = *(ListMap[0]);     
	*this *= ListCoef[0];
	CD_iCheckStatus;
	// compute the linear combination starting from second map
	for(long Idx=1;Idx < NbMap; Idx++) {
	  CD_iCheck( LinComb(*(ListMap[Idx]), ListCoef[Idx]) );}
      }
      CD_returnOK;
    }

   
 
  
  // Product term by term
  /////////////////////////////////////////////////
  /// Compute the term by term product MapIntern <- MapA*.MapIntern 
  /// \param MapA : map applied to the intern map
  /////////////////////////////////////////////////
  Status Prod(const MapExt<T> &MapA)
  {
    CD_iCheck( Conformable(MapA) );// check that Nside and Ordering of both maps match
    for(long idx=0; idx < this->npix_; idx++) 
      (*this)[idx] *= MapA[idx];
    CD_returnOK;
  }

  /////////////////////////////////////////////////
  /// Compute the term by term product MapIntern <- MapA*.MapIntern 
  /// \param MapA : map applied to the intern map
  /////////////////////////////////////////////////
  void operator *=(const MapExt<T> &MapA) {m_Status=Prod(MapA);}
  
  /////////////////////////////////////////////////
  /// Add term by term : MapIntern <- MapA + MapIntern 
  /// \param MapA : map applied to the intern map
  /////////////////////////////////////////////////
  void operator +=(const MapExt<T> &MapA)
  {
    CD_iCheck( Conformable(MapA) );// check that Nside and Ordering of both maps match
    for(long idx=0; idx < this->npix_; idx++) 
      (*this)[idx] += MapA[idx];
    CD_returnOK;
  }
  
  /////////////////////////////////////////////////
  /// Compute the sum of pixel 
  /////////////////////////////////////////////////
  T sum()
  { 
    T s = 0;
    for(long idx=0; idx < this->npix_; idx++)    
      s += ((*this)[idx]);
    return s;
  }
  
  
  /////////////////////////////////////////////////
  /// Compute the product MapIntern <- MapIntern*Coef 
  /// \param Coef : coefficent applied to all pixel of the intern map
  /////////////////////////////////////////////////
  template<typename T2> Status operator *=(const T2 &Coef) 
    {
      for(long idx=0; idx < this->npix_; idx++) 
	(*this)[idx] *= Coef;
      CD_returnOK;
    }
  
  /////////////////////////////////////////////////
  /// Compute the term by term product MapIntern <- Map1*  *MapNbMap
  /// if empty object, load it, compute and free it
  /// \param ListMap : collection of maps to multiplied
  /// \param NbMap : number of maps
  /////////////////////////////////////////////////
  Status SetWithProd(const  MapExt<T> **ListMap, long NbMap)
  {
    MapExt<T> TempMap; // temp map object used to load the different maps
    
    //check on number of maps
    if (NbMap > 0) {
      *this = *(ListMap[0]);     
      // compute the product starting from second map
      for(long Idx=1;Idx < NbMap; Idx++) {      
	(*this) *= *(ListMap[Idx]);  CD_iCheckStatus; //accumulate the product
      }
    }
    CD_returnOK;
  }

  // Pretty print
  /////////////////////////////////////////////////
  /// Print pixel value between two indexes
  /// \param begin : index of the first pixel to print
  /// \param end : index of the last pixel to print
  /// \param title : title of the map
  /////////////////////////////////////////////////
  void PrintMap(long begin, long end, const string &title="") ;

  /////////////////////////////////////////////////
  /// Print pixel value of the whole map
  /// \param title : title of the map
  /////////////////////////////////////////////////
  void PrintMap(const string &title="")  
  {PrintMap(0, this->npix_-1, title);}
  
  /////////////////////////////////////////////////
  /// Print attribut of class
  /// \param Name : title of the map
  /////////////////////////////////////////////////
  void PrintAttribut(const string &Name="") ;
  

// Check 
  /////////////////////////////////////////////////
  ///  Check that Nside and Ordering of internal and other map match
  ///  return 0 if OK
  /// \param Map : other map
  template <typename T2> Status Conformable(const MapExt<T2> &Map)
    {
    if (!Healpix_Map<T>::conformable(Map)){ CD_return(10); }
    CD_returnOK;
  }
  
  // Alm <- SHT -> Map : SHT : Spherical Harmonic Transformation
  /////////////////////////////////////////////////
  /// Create an Alm object computed from map using SHT 
  /// \param Lmax : maximum order of l
  /// \param Mmax : maximum order of m
  /////////////////////////////////////////////////
  AlmExt<T> * ToAlm(int Lmax=CD_LMAX, int Mmax=CD_MMAX)
    {
      // create a Alm object 
      AlmExt<T> *pRetAlm = new  AlmExt<T>(Lmax, Mmax); //create new object Alm
      
      if (pRetAlm == NULL) {
	CD_WarningStatus(7);
	return(NULL);        
      }
      // call Alm FromMap method with the current object map
      if (pRetAlm->FromMap(*this) != 0) {
	delete pRetAlm;
	CD_WarningStatus(7);
	return(NULL);  
      }
      return(pRetAlm); 
    }
  
    
  /////////////////////////////////////////////////
  /// Initialize map pixel values computing an iSHT from Alm
  /// \param Alm : spherical harmonic coefficient 
  /////////////////////////////////////////////////
  Status FromAlm(const AlmExt<T> &Alm)
  {
    //double offset = ObjAlm(0,0).real()/sqrt(fourpi);
    //ObjAlm(0,0) = 0;
    alm2map(Alm, *this);
    //(*this).add(offset);    
    CD_returnOK;
  }

  /////////////////////////////////////////////////
  /// Initialize map pixel values computing an iSHT from Alm
  /// \param AlmT : Temperature spherical harmonic coefficients 
  /// \param AlmE : E modes
  /// \param AlmB : B modes
  /// \param mapQ : Q polar map
  /// \param mapU : U polar map
  /////////////////////////////////////////////////
  Status FromPolarAlm(const AlmExt<T> &AlmT , const AlmExt<T> &AlmE,const AlmExt<T> &AlmB,MapExt<T> &mapQ, MapExt<T> &mapU)
  {
    alm2map_pol (AlmT, AlmE, AlmB, *this, mapQ, mapU);
    //double offset = ObjAlm(0,0).real()/sqrt(fourpi);
    //ObjAlm(0,0) = 0;
    //alm2map(Alm, *this);
    //(*this).add(offset);    
    CD_returnOK;
  }

  /////////////////////////////////////////////////
  /// Initialize map pixel values computing an iSHT from Alm
  /// \param AlmT : Temperature spherical harmonic coefficients 
  /// \param AlmE : E modes
  /// \param AlmB : B modes
  /// \param mapQ : Q polar map
  /// \param mapU : U polar map
  /// \param Nside : resolution of the map
  /// \param scheme : ordering of the map
  /////////////////////////////////////////////////
  Status FromPolarAlm(const AlmExt<T> &AlmT, const AlmExt<T> &AlmE, const AlmExt<T> &AlmB, MapExt<T> &mapQ, MapExt<T> &mapU, int Nside, 
		 Healpix_Ordering_Scheme scheme=RING) 
  { 
    mapQ.SetNside (Nside,scheme);
    mapU.SetNside (Nside,scheme);
    this->SetNside(Nside,scheme); CD_return(FromPolarAlm(AlmT, AlmE, AlmB, mapQ, mapU)); }
  

  /////////////////////////////////////////////////
  /// Initialize map pixel values computing an iSHT from Alm
  /// \param Alm : spherical harmonic coefficient 
  /// \param Nside : resolution of the map
  /// \param scheme : ordering of the map
  /////////////////////////////////////////////////
  Status FromAlm(const AlmExt<T> &Alm, int Nside, 
		 Healpix_Ordering_Scheme scheme=RING) 
  { this->SetNside(Nside,scheme); CD_return(FromAlm(Alm)); }
  
  /////////////////////////////////////////////////
  /// Returns the minimum and maximum value of the map in Min and Max.
  /// \param Min : minimum value to be set
  /// \param Max : maximim value to be set
  /////////////////////////////////////////////////
  Status minmax (T &Min, T &Max)
  {
    Min = 1e30; Max = -1e30;
    for (int m=0; m<this->Npix(); ++m)
      {
	T val = (*this)[m];
	if (!approx<double>(val,Healpix_undef))
	  {
	    if (val>Max) Max=val;
	    if (val<Min) Min=val;
	  }
      }
  }

  
  

protected :

  /// initialisation data member or attribut of object
  void InitAttribut() {m_Status=0; }
  
  /// Identify error of last call method, O is Ok
  Status m_Status;
  
};



//#include "MapExt.cpp"
//
// Template class => define methods directely in header
// 

#endif
