/*
 * Project: nilc
 * Date: 11/01  
 * Authors: M. Le Jeune
 *
 */

/* Copyright (C) 2011 APC CNRS Université Paris Diderot <lejeune@apc.univ-paris7.fr>  
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see http://www.gnu.org/licenses/gpl.html
 */

#include "fits_tool.h"
#include <iostream>
#include <string>
#include "paramfile.h"
#include "processoptions.h"
#include <sys/time.h>

int main(int argc ,const char** argv)
{
  srand ( time(NULL) );
  timeval start, end; 
  char** ptr ;
  ptr = const_cast <char **> (argv);
  
  /// options
  bool almopt=false; // input are alm files
  bool pixel_based=false; // pixel domain localisation only

  /// inputs
  string outfile; // healpix map FITS file
  string *infile; // healpix map FITS file
  int nbin;

  /// parameters
  arr<double> mixcol;   // size is nbin
  arr<int> bands;     // not used
  string *zonefile; // healpix map FITS file
  int nbands;
  int lmax; 
  int nside  = -1;
  int n_iter = 0;

  
  // command line parsing (1/2)
  string longopt[] = {"almopt", "n_iter", "nside", "pixel_based", "zonefile"};
  ArgType type[]   = {FLAG, INT, INT, FLAG, STRING};
  string usage[]   = {"Inputs are alm FITS file.", "Number of iteration of anafast.", "Resolution of the output map.", "Pixel based analysis", "Pixel zone definition"};
  char shortopt[]  = "ainpz";
  Args args = Args(5, longopt, shortopt, type, usage,"Perform needlet ILC.", "outfile l0 J zonefile1 lmax1 ... zonefileJ lmaxJ N infile1 coeff1 ... infileN coeffN \n -p outfile -z zonefile N infile1 coeff1 ... infileN coeffN ");
  
  
  // parameter file
  if ((argc == 2) && (argv[1][0]!='-'))
    {
      module_startup ("nilc", argc, argv, 2, "<parameter file>");
      paramfile params (argv[1]);
      almopt = params.find<bool> ("almopt",false);
      nside  = params.find<int>  ("nside" ,-1);
      n_iter = params.find<int>  ("n_iter",0);
      outfile= params.find<string> ("outfile","");
      string sinfile = params.find<string> ("infile","");
      nbin = parse_list (sinfile, &infile);
      string smixcol = params.find<string> ("mixcol","");
      string*stmixcol;
      int nbin2 = parse_list (smixcol, &stmixcol);
      if (nbin!=nbin2){
	fprintf(stderr,"\nERROR!!! Can't parse 'mixcol' parameter. Found %d elements vs %d inputs\n",nbin2,nbin);
	return 1;
      }
      mixcol.alloc (nbin);
      for (int i=0; i<nbin; i++)
	mixcol[i] = atof(stmixcol[i].c_str());
      delete []stmixcol;

      string szonefile = params.find<string> ("zonefile","");
      nbands = parse_list (szonefile, &zonefile);
      string sbands = params.find<string> ("bands","");
      string* stbands;
      int nbands2 = parse_list (sbands, &stbands);
      if (nbands2 == 0)	
	pixel_based = true;
      else if (nbands!=nbands2-2){
	fprintf(stderr,"\nERROR!!! Can't parse 'bands' parameter. Found %d elements vs %d zones\n",nbands2,nbands);
	return 1;
      }
      else {
	bands.alloc (nbands2);
	for (int i=0; i<nbands2; i++)
	  bands[i] = atoi(stbands[i].c_str());
	delete []stbands;
      }
    }
  
  // command line parsing (2/2)
  else    {
    args.processargs(argc, ptr);
    args.info();    //args.printusage();
    almopt  = args["almopt"];
    args.get(n_iter,"n_iter");
    args.get(nside,"nside");
    pixel_based  = args["pixel_based"];
    if (pixel_based){
      string zfile = "";
      args.get(zfile,"zonefile");
      if (zfile.size()>1){
	zonefile = new string[1]; 
	zonefile[0] = zfile;
	nbands = 1;}
      else
	nbands = 0;
      int idx=0;
      nbin = atoi(args[idx++].c_str());
      infile = new string[nbin]; 
      mixcol.alloc(nbin);
      for (int i=0; i<nbin; i++){
	infile[i] = args[idx++];
	mixcol[i] = atof(args[idx++].c_str());}
    }
    else{
      int idx = 1;
      outfile = args[idx++];
      int l0  = atoi(args[idx++].c_str());
      nbands  = atoi(args[idx++].c_str());
      bands.alloc(nbands+2);
      zonefile = new string[nbands]; 
      bands[0] = 0;
      bands[1] = l0;
      for (int i=0; i<nbands; i++){
	zonefile[i] = args[idx++];
	bands[i+2]  = atoi(args[idx++].c_str());}
      nbin = atoi(args[idx++].c_str());
      infile = new string[nbin]; 
      mixcol.alloc(nbin);
      for (int i=0; i<nbin; i++){
	infile[i] = args[idx++];
	mixcol[i] = atof(args[idx++].c_str());}
    }
  }
    
  // preprocessing
  fitshandle inp;
  if (!almopt){
    inp.open(infile[0]);
    inp.goto_hdu(2);
    nside = (int) sqrt(inp.nelems(1)/12);
  }
  if (nside<0){
    fprintf(stderr,"\nERROR!!! Missing 'nside' parameter\n");
    return 1;
  }
  
  bool ztorm=false;
  if (nbands==0){
    get_lst_tmp (1, "zone", &zonefile);
    MapExt<int> mask;
    mask.SetNside(nside, RING);
    mask.fill(1);
    fitshandle out;
    out.create(zonefile[0]);
    write_Healpix_map_to_fits (out, mask,planckType<int>());
    ztorm=true;
  }
  arr<double> onearr(nbands);
  onearr.fill(1);
  arr<int> nsides(nbands);
  for (int j=0; j<nbands; j++)    {
    inp.open(zonefile[j]);
    inp.goto_hdu(2);
    nsides[j] = (int) sqrt(inp.nelems(1)/12);
  }
  

  // pixel based case
  if (pixel_based){
    string statfile = get_tmp_file("stat");
    string ilcfile  = get_tmp_file("ilc");
    fprintf(stdout, "Compute second order stats\n");
    if (almopt){
      string* mapfile;
      get_lst_tmp (nbin, "map", &mapfile);
      for (int m=0; m<nbin; m++)
	alm2map<double>  (infile[m], mapfile[m], nside, false);
      infile = mapfile;}
    map2stat<double> (infile, nbin, statfile, zonefile[0]);
    fprintf(stdout, "Compute ILC filter\n");
    stat2ILCfilter<double> (statfile, nbin, ilcfile, mixcol);
    fprintf(stdout, "Apply ILC filter\n ");
    combinezone<double> (infile, nbin, outfile, ilcfile,zonefile[0]);
    rm_tmp_file(statfile);
    rm_tmp_file(ilcfile);
    if (ztorm)
      rm_lst_tmp(1, &zonefile);
  }
  
  else{
    string *winfile;
    get_lst_tmp (nbands, "window", &winfile);
    string* scalefile;
    get_lst_tmp (nbands, "cfalm", &scalefile);
    makesplinefilters (bands, winfile, 0);
    for (int j=0; j<nbands; j++) {
      string* mapfile;
      get_lst_tmp (nbin, "needlet", &mapfile);
      
      for (int m=0; m<nbin; m++) {
	fprintf(stdout, "Perform analysis of map %d and scale %d. Lmax is %d, Nside is %d\n ", m, j, bands[j+2], nsides[j]);
	if (almopt)
	  alm2need<double> (infile[m], mapfile[m], bands[j+2], nsides[j], winfile[j]);
	else
	  map2need<double> (infile[m], mapfile[m], bands[j+2], nsides[j], n_iter,winfile[j]);
      }
      string statfile = get_tmp_file("stat");
      string ilcfile  = get_tmp_file("ilc");
      string cmapfile = get_tmp_file("cmap");
      fprintf(stdout, "Compute ILC filter for scale %d\n ", j);
      map2stat<double> (mapfile, nbin, statfile, zonefile[j]);
      stat2ILCfilter<double> (statfile, nbin, ilcfile, mixcol);
      fprintf(stdout, "Apply ILC filter for scale %d\n ", j);
      combinezone<double> (mapfile, nbin, cmapfile, ilcfile,zonefile[j]);
      need2alm<double> (cmapfile, scalefile[j], bands[j+2], n_iter, winfile[j]);
      rm_tmp_file(statfile);
      rm_tmp_file(ilcfile);
      rm_tmp_file(cmapfile);
      rm_lst_tmp(nbin, &mapfile);
    }
    string almfile = get_tmp_file("alm");
    fprintf(stdout, "Combine scales\n ");
    combinealm<double>(scalefile, nbands, almfile, onearr, false);
    alm2map<double> (almfile, outfile, nside, 0);
    rm_lst_tmp(nbands ,&scalefile);
    rm_lst_tmp(nbands, &winfile);
    rm_tmp_file(almfile);
  }
  
  return 0;
  
}
